//
// A copyright notice must contain a "C" enclosed in parentheses: (C) 
//

package expandedjavaparser.Ast;

import lpg.lpgjavaruntime.*;

/**
 *<b>
 *<li>Rule 192:  SynchronizedStatement ::= synchronized ( Expression ) Block
 *</b>
 */
public class SynchronizedStatement extends Ast implements ISynchronizedStatement
{
    private IExpression _Expression;
    private Block _Block;

    public IExpression getExpression() { return _Expression; }
    public Block getBlock() { return _Block; }

    public SynchronizedStatement(IToken leftIToken, IToken rightIToken,
                                 IExpression _Expression,
                                 Block _Block)
    {
        super(leftIToken, rightIToken);

        this._Expression = _Expression;
        this._Block = _Block;
        initialize();
    }

    public boolean equals(Object o)
    {
        if (o == this) return true;
        //
        // The supers call is not required for now because Ast nodes
        // can only extend the root Ast, AstToken and AstList and none
        // of these nodes contain children.
        //
        // if (! super.equals(o)) return false;
        //
        if (! (o instanceof SynchronizedStatement)) return false;
        SynchronizedStatement other = (SynchronizedStatement) o;
        if (! _Expression.equals(other.getExpression())) return false;
        if (! _Block.equals(other.getBlock())) return false;
        return true;
    }

    public int hashCode()
    {
        int hash = 7;
        hash = hash * 31 + (getExpression().hashCode());
        hash = hash * 31 + (getBlock().hashCode());
        return hash;
    }

    public void accept(Visitor v) { v.visit(this); }
    public void accept(ArgumentVisitor v, Object o) { v.visit(this, o); }
    public Object accept(ResultVisitor v) { return v.visit(this); }
    public Object accept(ResultArgumentVisitor v, Object o) { return v.visit(this, o); }
}


