#include "control.h"
#include "option.h"
#include <errno.h>

#include <iostream>
using namespace std;

const char *Option::default_block_begin = "/.",
           *Option::default_block_end = "./";

//
//
//
const char *Option::AllocateString(const char *str_)
{
    char *str = new char[strlen(str_) + 1];
    strcpy(str, str_);

    return str;
}

const char *Option::AllocateString(const char *str_, char c)
{
    int length = strlen(str_) + 4;
    char *str = new char[length + 1];
    strcpy(str, str_);
    str[length - 4] = '=';
    str[length - 3] = '\'';
    str[length - 2] = c;
    str[length - 1] = '\'';
    str[length] = '\0';

    return str;
}

const char *Option::AllocateString(const char *str_, const char *str2)
{
    char *str = new char[strlen(str_) + strlen(str2) + 4];
    strcpy(str, str_);
    strcat(str, "=");
    if (*str2 != '(')
        strcat(str, "\"");
    strcat(str, str2);
    if (*str2 != '(')
        strcat(str, "\"");

    return str;
}

const char *Option::AllocateString(const char *str_, int i)
{
    char *str = new char[strlen(str_) + 13 + 1];
    IntToString num(i);
    strcpy(str, str_);
    strcat(str, "=");
    strcat(str, num.String());

    return str;
}

//
// action_block
// ast_directory
// ast_type
// automatic_ast
// attributes
//
const char *Option::ClassifyA(const char *start, bool flag)
{
    if (flag)
    {
        int i = OptionMatch(start, "action", "block");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (q == NULL)
            {
                return (i > 0
                          ? ReportMissingValue(start, "ACTION_BLOCK")
                          : ReportAmbiguousOption(start, "ACTION_BLOCK, AST_DIRECTORY, AST_TYPE, ATTRIBUTES, AUTOMATIC_AST"));
            }
            else
            {
                if (*q == '(')
                {
                    const char *filename;
                    p = GetStringValue("Filename in ACTION_BLOCK", q + 1, filename);
                    if (*filename == '*')
                    {
                        char *name = new char[strlen(filename) + strlen(file_prefix)];
                        strcpy(name, file_prefix);
                        strcat(name, filename + 1);
                        filename = name;
                        temp_string.Next() = name;
                    }

                    const char *block_begin;
                    p = GetStringValue("Block Begin Marker in ACTION_BLOCK", CleanUp(p), block_begin);
                    const char *block_end;
                    p = GetStringValue("Block End Marker in ACTION_BLOCK", CleanUp(p), block_end);
                    p = CleanUp(p);
                    if (*p == ')')
                    {
                        ProcessBlock(filename, block_begin, block_end);
                        return ++p;
                    }

                    report.Put("***ERROR: Illegal ACTION_BLOCK option specified: ");
                    while (start < p)
                         report.PutChar(*start++);
                    report.Put(". A value of the form \"(filename,begin_block_marker,end_block_marker)\" was expected.\n");
                    return_code = 12;

                    return p;
                }
            }
        }

        i = OptionMatch(start, "ast", "directory");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (q == NULL)
            {
                return (i > 3
                          ? ReportMissingValue(start, "AST_DIRECTORY")
                          : ReportAmbiguousOption(start, "AST_DIRECTORY, AST_TYPE"));
            }
            else p = GetStringValue("AST_DIRECTORY", q, ast_directory);

            return (i < 4
                      ? ReportAmbiguousOption(start, "AST_DIRECTORY, AST_TYPE")
                      : p);
        }

        i = OptionMatch(start, "ast", "type");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (q == NULL)
                 return ReportMissingValue(start, "AST_TYPE");
            else p = GetStringValue("AST_TYPE", q, ast_type);

            return p;
        }
    }

    int i = strxsub(start + 1, "ttributes");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        attributes = flag;
        const char *p = start + i + 1;
        return (ValuedOption(p) ? ReportValueNotRequired(start, "ATTRIBUTES") : p);
    }

    i = OptionMatch(start, "automatic", "ast");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        const char *p = start + i + 1,
                   *q = ValuedOption(p);

        if (q == NULL)
        {
            automatic_ast = (flag ? NESTED : NONE);
            return p;
        }
        else if (flag) // Cannot assign a value to "noautomatic_ast"
        {
            const char *value;
            p = GetStringValue("AUTOMATIC_AST", q, value);
            int length = strlen(value);
            if (strxsub(value, "none") == length)
                 automatic_ast = NONE;
            else if (strxsub(value, "nested") == length)
                 automatic_ast = NESTED;
            else if (strxsub(value, "toplevel") == length)
                 automatic_ast = TOPLEVEL;
            else
            {
                report.Put("***ERROR: \"");
                report.Put(value);
                report.Put("\" is an invalid value for option \"");
                for (int k = 0; k < i + 1; k++)
                    report.PutChar(start[k]);
                report.Put("\"\n");
                return_code = 12;
            }

            return p;
        }
    }

    return ClassifyBadOption(start, flag);
}

//
// backtrack
// byte
//
const char *Option::ClassifyB(const char *start, bool flag)
{
    int i = strxsub(start + 1, "acktrack");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        backtrack = flag;
        const char *p = start + i + 1;
        return (i < 1 ? ReportAmbiguousOption(start, "BACKTRACK, BYTE")
                      : ValuedOption(p) ? ReportValueNotRequired(start, "BACKTRACK") : p);
    }

    i = strxsub(start + 1, "yte");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        byte = flag;
        const char *p = start + i + 1;
        return (ValuedOption(p) ? ReportValueNotRequired(start, "BYTE") : p);
    }

    return ClassifyBadOption(start, flag);
}

//
// conflicts
//
const char *Option::ClassifyC(const char *start, bool flag)
{
    int i = strxsub(start + 1, "onflicts");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        conflicts = flag;
        const char *p = start + i + 1;
        return (ValuedOption(p) ? ReportValueNotRequired(start, "CONFLICTS") : p);
    }

    return ClassifyBadOption(start, flag);
}

//
// dat_directory
// dat_file
// dcl_file
// def_file
// debug
//
const char *Option::ClassifyD(const char *start, bool flag)
{
    if (flag)
    {
        int i = OptionMatch(start, "dat", "directory");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (q == NULL)
            {
                return (i > 0
                          ? ReportMissingValue(start, "DAT_DIRECTORY")
                          : ReportAmbiguousOption(start, "DAT_DIRECTORY, DAT_FILE, DCL_FILE, DEF_FILE, DEBUG"));
            }
            else
            {
                p = GetStringValue("DAT_DIRECTORY", q, dat_directory);
                return (i < 1
                          ? ReportAmbiguousOption(start, "DAT_DIRECTORY, DAT_FILE, DCL_FILE, DEF_FILE")
                          : i < 4
                              ? ReportAmbiguousOption(start, "DAT_DIRECTORY, DAT_FILE")
                              : p);
            }
        }

        i = OptionMatch(start, "dat", "file");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (q == NULL)
            {
                if (i == 1 && (start[1] == 'f' || start[1] == 'F'))
                     return ReportMissingValue(start, "DAT_FILE, DCL_FILE, DEF_FILE");
                else if (i > 0)
                     return ReportMissingValue(start, "DAT_FILE");
            }
            else
            {
                p = GetStringValue("DAT_FILE", q, dat_file);
                return ((i == 1 && (start[1] == 'f' || start[1] == 'F'))
                                ? ReportAmbiguousOption(start, "DAT_FILE, DCL_FILE, DEF_FILE")
                                : p);
            }
        }

        i = OptionMatch(start, "dcl", "file");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (q == NULL)
            {
                if (i > 0)
                     return ReportMissingValue(start, "DCL_FILE");
            }
            else
            {
                p = GetStringValue("DCL_FILE", q, dcl_file);
                return p;
            }
        }

        i = OptionMatch(start, "def", "file");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (q == NULL) // Must have a value
                return ReportMissingValue(start, "DEF_FILE");
            else
            {
                p = GetStringValue("DEF_FILE", q, def_file);
                return p;
            }
        }
    }

    int i = strxsub(start + 1, "ebug");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        debug = flag;
        const char *p = start + i + 1;
        return (ValuedOption(p) ? ReportValueNotRequired(start, "DEBUG") : p);
    }

    return ClassifyBadOption(start, flag);
}

//
// edit
// error_maps
// escape
// export_terminals
// extends_parsetable
//
const char *Option::ClassifyE(const char *start, bool flag)
{
    int i = OptionMatch(start, "extends", "parsetable");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        const char *p = start + i + 1,
                   *q = ValuedOption(p);
        if (i < 1)
             return ReportAmbiguousOption(start, "EDIT, ERROR-MAPS, ESCAPE, EXPORT-TERMINALS, EXTENDS_PARSETABLE");
        else if (i < 2)
             return ReportAmbiguousOption(start, "EXPORT-TERMINALS, EXTENDS_PARSETABLE");
        else if (q == NULL && flag)
        {
            char *temp = new char[1];
            *temp = NULL_CHAR;
            extends_parsetable = temp;
            temp_string.Next() = extends_parsetable;
            return p;
        }
        else if (flag) // Cannot assign a value to "noextends_parsetable"
        {
            p = GetStringValue("EXTENDS-PARSETABLE", q, extends_parsetable);
            return p;
        }
    }

    if (flag)
    {
        int i = strxsub(start + 1, "scape");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (i < 1)
                return ReportAmbiguousOption(start, "ESCAPE, EXPORT_TERMINALS");
            else if (q == NULL) // Must have a value
                return ReportMissingValue(start, "ESCAPE");
            else
            {
                const char *escape;
                p = GetStringValue("ESCAPE", q, escape);
                if (escape[1] != NULL_CHAR) // more than a single character?
                {
                    report.Put("***ERROR: \"");
                    report.Put(escape);
                    report.Put("\" is an invalid value for option \"");
                    for (int k = 0; k < i + 1; k++)
                        report.PutChar(start[k]);
                    report.Put("\"\n");
                    return_code = 12;
                }
                this -> escape = *escape;

                return p;
            }
        }

        i = OptionMatch(start, "export", "terminals");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (q == NULL)
            {
                if (i > 0)
                     return ReportMissingValue(start, "EXPORT_TERMINALS");
            }
            else
            {
                if (*q != '(') // just the file name is provided.
                {
                    p = GetStringValue("EXPORT_TERMINALS", q, exp_file);
                    return p;
                }
                else // a parenthesized list of 1, 2 or three arguments: (filename [, prefix] [, suffix])
                {
                    p = GetStringValue("Filename in EXPORT_TERMINALS", q + 1, exp_file);
                    if (*exp_file == '*')
                    {
                        char *name = new char[strlen(exp_file) + strlen(file_prefix)];
                        strcpy(name, file_prefix);
                        strcat(name, exp_file + 1);
                        exp_file = name;
                        temp_string.Next() = name;
                    }

                    p = CleanUp(p);
                    if (*p == ')')
                        return ++p;

                    p = GetStringValue("Prefix in EXPORT_TERMINALS", CleanUp(p), exp_prefix);
                    p = CleanUp(p);
                    if (*p == ')')
                        return ++p;

                    p = GetStringValue("Suffix in EXPORT_TERMINALS", CleanUp(p), exp_suffix);
                    p = CleanUp(p);
                    if (*p == ')')
                        return ++p;

                    report.Put("***ERROR: Illegal EXPORT_TERMINALS option specified: ");
                    while (start < p)
                         report.PutChar(*start++);
                    report.Put(". A value of the form \"(filename [,prefix] [,suffix])\" was expected.\n");
                    return_code = 12;

                    return p;
                }
            }
        }
    }

    i = OptionMatch(start, "error", "maps");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        error_maps = flag;
        const char *p = start + i + 1;
        return (i < 1 ? ReportAmbiguousOption(start, "EDIT, ERROR_MAPS")
                      : ValuedOption(p) ? ReportValueNotRequired(start, "ERROR_MAPS") : p);
    }

    i = strxsub(start + 1, "dit");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        edit = flag;
        const char *p = start + i + 1;
        return (ValuedOption(p) ? ReportValueNotRequired(start, "EDIT") : p);
    }

    return ClassifyBadOption(start, flag);
}

//
// factory
// file_prefix
// filter
// first
// follow
//
const char *Option::ClassifyF(const char *start, bool flag)
{
    if (flag)
    {
        int i = strxsub(start + 1, "actory");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {

            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (q == NULL) // Must have a value
            {
                return (i < 1
                          ? ReportAmbiguousOption(start, "FACTORY, FILE_PREFIX, FILTER, FIRST, FOLLOW")
                          : (i == 1 && (start[1] == 'i' || start[1] == 'I'))
                                ? ReportAmbiguousOption(start, "FILE_PREFIX, FILTER, FIRST")
                                : ReportMissingValue(start, "FILE_PREFIX"));
            }
            else
            {
                p = GetStringValue("FACTORY", q, factory);
                return p;
            }
        }

        i = OptionMatch(start, "file", "prefix");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (q == NULL) // Must have a value
            {
                if (i > 0)
                     return (i == 1 && (start[1] == 'i' || start[1] == 'I'))
                                ? ReportAmbiguousOption(start, "FILE_PREFIX, FILTER, FIRST")
                                : ReportMissingValue(start, "FILE_PREFIX");
            }
            else
            {
                p = GetStringValue("FILE_PREFIX", q, file_prefix);
                return p;
            }
        }

        i = strxsub(start + 1, "ilter");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {

            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (q == NULL) // Must have a value
                return ReportMissingValue(start, "FILTER");
            else
            {
                p = GetStringValue("FILTER", q, filter);
                ProcessPath(filter_file, filter);
                return p;
            }
        }
    }

    int i = strxsub(start + 1, "irst");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        first = flag;
        const char *p = start + i + 1;
        return (i < 1 ? ReportAmbiguousOption(start, "FIRST, FOLLOW")
                      : ValuedOption(p) ? ReportValueNotRequired(start, "FIRST") : p);
    }

    i = strxsub(start + 1, "ollow");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        follow = flag;
        const char *p = start + i + 1;
        return (ValuedOption(p) ? ReportValueNotRequired(start, "FOLLOW") : p);
    }

    return ClassifyBadOption(start, flag);
}

//
// goto_default
//
const char *Option::ClassifyG(const char *start, bool flag)
{
    int i = OptionMatch(start, "goto", "default");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        goto_default = flag;
        const char *p = start + i + 1;
        return (ValuedOption(p) ? ReportValueNotRequired(start, "GOTO_DEFAULT") : p);
    }

    return ClassifyBadOption(start, flag);
}


//
// Headers
//
const char *Option::ClassifyH(const char *start, bool flag)
{
    if (flag)
    {
        int i = strxsub(start + 1, "eaders");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (q == NULL)
            {
                if (i > 0)
                    return ReportMissingValue(start, "HEADERS");
            }
            else
            {
                if (*q == '(')
                {
                    const char *filename;
                    p = GetStringValue("Filename in HEADERS", q + 1, filename);
                    if (*filename == '*')
                    {
                        char *name = new char[strlen(filename) + strlen(file_prefix)];
                        strcpy(name, file_prefix);
                        strcat(name, filename + 1);
                        filename = name;
                        temp_string.Next() = name;
                    }

                    const char *block_begin;
                    p = GetStringValue("Block Begin Marker in HEADERS", CleanUp(p), block_begin);
                    const char *block_end;
                    p = GetStringValue("Block End Marker in HEADERS", CleanUp(p), block_end);
                    p = CleanUp(p);

                    if (*p == ')')
                    {
                        ProcessHeader(filename, block_begin, block_end);
                        return ++p;
                    }

                    report.Put("***ERROR: Illegal HEADERS option specified: ");
                    while (start < p)
                         report.PutChar(*start++);
                    report.Put(". A value of the form \"(filename,begin_block_marker,end_block_marker)\" was expected.\n");
                    return_code = 12;

                    return p;
                }
            }
        }
    }

    return ClassifyBadOption(start, flag);
}

//
// imp_file
// import_terminals
// include_directory/include_directories
//
const char *Option::ClassifyI(const char *start, bool flag)
{
    if (flag)
    {
        int i = OptionMatch(start, "imp", "file");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (i < 1)
                 return ReportAmbiguousOption(start, "IMP_FILE, IMPORT_TERMINALS, INCLUDE_DIRECTORY");
            else if (i < 3)
                 return ReportAmbiguousOption(start, "IMP_FILE, IMPORT_TERMINALS");
            else if (q == NULL) // Must have a value
                 return ReportMissingValue(start, "IMP_FILE");
            else
            {
                p = GetStringValue("IMP_FILE", q, imp_file);
                return p;
            }
        }

        i = OptionMatch(start, "import", "terminals");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (q == NULL) // Must have a value
                return ReportMissingValue(start, "IMPORT_TERMINALS");
            else
            {
                p = GetStringValue("IMPORT_TERMINALS", q, import_terminals);
                ProcessPath(import_file, import_terminals);
                return p;
            }
        }

        i = OptionMatch(start, "include", "directory");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (q == NULL) // Must have a value
                 return ReportMissingValue(start, "INCLUDE_DIRECTORY");
            else
            {
                p = GetStringValue("INCLUDE_DIRECTORY", q, include_directory);
                ProcessPath(include_search_directory, include_directory, "");
                return p;
            }
        }

        i = OptionMatch(start, "include", "directories"); // accept "directories" as a substitute for "directory".
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (q == NULL) // Must have a value
                 return ReportMissingValue(start, "INCLUDE_DIRECTORIES");
            else
            {
                p = GetStringValue("INCLUDE_DIRECTORIES", q, include_directory);
                return p;
            }
        }
    }

    return ClassifyBadOption(start, flag);
}

//
// lalr_level
// list
//
const char *Option::ClassifyL(const char *start, bool flag)
{
    int i = strxsub(start + 1, "alr");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        const char *p = start + i + 1,
                   *q = ValuedOption(p);
        if (q == NULL)
        {
            if (i > 0)
            {
                slr = ! flag;
                lalr_level = 1;
            }
            else ReportAmbiguousOption((flag ? start : start - 2), "LALR, LIST");
        }
        else if (flag) // Cannot assign a value to "nolalr"
        {
            const char *value;
            p = GetValue(q, value);

            slr = false;
            if (verify(value))
                lalr_level = atoi(value);
            else
            {
                report.Put("***ERROR: \"");
                report.Put(value);
                report.Put("\" is an invalid value for option \"");
                for (int k = 0; k < i + 1; k++)
                    report.PutChar(start[k]);
                report.Put("\"\n");
                return_code = 12;
            }

            return p;
        }
    }

    i = strxsub(start + 1, "ist");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        list = flag;
        const char *p = start + i + 1;
        return (ValuedOption(p) ? ReportValueNotRequired(start, "LIST") : p);
    }

    return ClassifyBadOption(start, flag);
}

//
// margin
// max_cases
//
const char *Option::ClassifyM(const char *start, bool flag)
{
    if (flag)
    {
        int i = strxsub(start + 1, "argin");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (q == NULL) // Must have a value
            {
                if (i > 0)
                     return ReportMissingValue(start, "MARGIN");
            }
            else
            {
                const char *value;
                p = GetValue(q, value);

                if (verify(value))
                    margin = atoi(value);
                else
                {
                    report.Put("***ERROR: \"");
                    report.Put(value);
                    report.Put("\" is an invalid value for option \"");
                    for (int k = 0; k < i + 1; k++)
                        report.PutChar(start[k]);
                    report.Put("\"\n");
                    return_code = 12;
                }

                return p;
            }
        }

        i = OptionMatch(start, "max", "cases");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (q == NULL) // Must have a value
            {
                if (i > 0)
                     return ReportMissingValue(start, "MAX_CASES");
            }
            else
            {
                const char *value;
                p = GetValue(q, value);

                if (verify(value))
                    max_cases = atoi(value);
                else
                {
                    report.Put("***ERROR: \"");
                    report.Put(value);
                    report.Put("\" is an invalid value for option \"");
                    for (int k = 0; k < i + 1; k++)
                        report.PutChar(start[k]);
                    report.Put("\"\n");
                    return_code = 12;
                }

                return p;
            }
        }
    }

    return ClassifyBadOption(start, flag);
}

//
// names
// nt_check
//
const char *Option::ClassifyN(const char *start, bool flag)
{
    if (flag)
    {
        int i = strxsub(start + 1, "ames");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (q == NULL) // Must have a value
            {
                if (i > 0)
                     return ReportMissingValue(start, "NAMES");
            }
            else
            {
                const char *value;
                p = GetStringValue("NAMES", q, value);
                int length = strlen(value);

                if (strxsub(value, "optimized") == length)
                    names = OPTIMIZED;
                else
                {
                    i = strxsub(value, "maximum");
                    if (length == 1)
                         return ReportAmbiguousOption(start, "MAXIMUM, MINIMUM");
                    else if (i == length)
                        names = MAXIMUM;
                    else if (strxsub(value, "minimum") == length)
                        names = MINIMUM;
                    else
                    {
                        report.Put("***ERROR: \"");
                        report.Put(value);
                        report.Put("\" is an invalid value for option \"");
                        for (int k = 0; k < i + 1; k++)
                            report.PutChar(start[k]);
                        report.Put("\"\n");
                        return_code = 12;
                    }
                }

                return p;
            }
        }
    }

    int i = OptionMatch(start, "nt", "check");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        nt_check = flag;
        const char *p = start + i + 1;
        return (ValuedOption(p) ? ReportValueNotRequired(start, "NT_CHECK") : p);
    }

    return ClassifyBadOption(start, flag);
}

//
// or_marker
//
const char *Option::ClassifyO(const char *start, bool flag)
{
    if (flag)
    {
        int i = OptionMatch(start, "or", "marker");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (q == NULL) // Must have a value
                 return ReportMissingValue(start, "OR_MARKER");
            else
            {
                const char *or_marker;
                p = GetStringValue("OR_MARKER", q, or_marker);
                if (or_marker[1] != NULL_CHAR) // more than a single character?
                {
                    report.Put("***ERROR: \"");
                    report.Put(or_marker);
                    report.Put("\" is an invalid value for option \"");
                    for (int k = 0; k < i + 1; k++)
                        report.PutChar(start[k]);
                    report.Put("\"\n");
                    return_code = 12;
                }
                this -> or_marker = *or_marker;

                return p;
            }
        }
    }

    return ClassifyBadOption(start, flag);
}

//
// parent_saved
// package
// parsetable_interfaces
// prefix
// priority
// programming_language
// prs_file
//
const char *Option::ClassifyP(const char *start, bool flag)
{
    // int i = OptionMatch(start, "parsetable", "interfaces");
    // if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    // {
    //     const char *p = start + i + 1,
    //                *q = ValuedOption(p);
    //     if (i < 1)
    //          return ReportAmbiguousOption(start, "PACKAGE, PARSETABLE-INTERFACES, PREFIX, PRIORITY, PROGRAMMING_LANGUAGE, PRS_FILE");
    //     else if (i < 2)
    //          return ReportAmbiguousOption(start, "PACKAGE, PARSETABLE-INTERFACES");
    //     else if (q == NULL && flag)
    //     {
    //         char *temp = new char[1];
    //         *temp = NULL_CHAR;
    //         parsetable_interfaces = temp;
    //         temp_string.Next() = parsetable_interfaces;
    //         return p;
    //     }
    //     else if (flag) // Cannot assign a value to "noparsetable_interfaces"
    //     {
    //         p = GetStringValue("PARSETABLE-INTERFACES", q, parsetable_interfaces);
    //         return p;
    //     }
    // }

    if (flag)
    {
        int i = strxsub(start + 1, "ackage");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (i < 1)
                 return ReportAmbiguousOption(start, "PACKAGE, PARENT_SAVED, PARSETABLE-INTERFACES, PREFIX, PRIORITY, PROGRAMMING_LANGUAGE, PRS_FILE");
            else if (i < 2)
                 return ReportAmbiguousOption(start, "PACKAGE, PARENT_SAVED, PARSETABLE-INTERFACES");
            else if (q == NULL) // Must have a value
                 return ReportMissingValue(start, "PACKAGE");
            else
            {
                p = GetStringValue("PACKAGE", q, package);
                return p;
            }
        }

        i = OptionMatch(start, "parsetable", "interfaces");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (i < 3)
                 return ReportAmbiguousOption(start, "PARENT_SAVED, PARSETABLE-INTERFACES");
            if (q == NULL && flag)
            {
                char *temp = new char[1];
                *temp = NULL_CHAR;
                parsetable_interfaces = temp;
                temp_string.Next() = parsetable_interfaces;
                return p;
            }
            else if (flag) // Cannot assign a value to "noparsetable_interfaces"
            {
                p = GetStringValue("PARSETABLE-INTERFACES", q, parsetable_interfaces);
                return p;
            }
        }

        i = strxsub(start + 1, "refix");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (i < 2)
                 return ReportAmbiguousOption(start, "PREFIX, PRIORITY, PROGRAMMING_LANGUAGE, PRS_FILE");
            else if (q == NULL) // Must have a value
                 return ReportMissingValue(start, "PREFIX");
            else
            {
                p = GetStringValue("PREFIX", q, prefix);
                return p;
            }
        }

        i = strxsub(start + 1, "riority");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            priority = flag;
            const char *p = start + i + 1;
            return (ValuedOption(p) ? ReportValueNotRequired(start, "PRIORITY") : p);
        }

        i = OptionMatch(start, "programming", "language");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p),
                       *value;
            if (q == NULL) // Must have a value
                return ReportMissingValue(start, "PROGRAMMING_LANGUAGE");
            else
            {
                p = GetStringValue("PROGRAMMING_LANGUAGE", q, value);
                int length = strlen(value);
                if (strxsub(value, "none") == length ||
                    strxsub(value, "xml") == length)
                     programming_language = XML;
                else if (strxsub(value, "c") == length)
                     programming_language = C;
                else if (strxsub(value, "cpp") == length || strxsub(value, "c++") == length)
                     programming_language = CPP;
                else if (strxsub(value, "java") == length)
                     programming_language = JAVA;
                else if (strxsub(value, "plx") == length)
                     programming_language = PLX;
                else if (strxsub(value, "plxasm") == length)
                     programming_language = PLXASM;
                else if (strxsub(value, "ml") == length)
                     programming_language = ML;
                else
                {
                    report.Put("***ERROR: \"");
                    report.Put(value);
                    report.Put("\" is an invalid value for option \"");
                    for (int k = 0; k < i + 1; k++)
                        report.PutChar(start[k]);
                    report.Put("\"\n");
                    return_code = 12;
                }

                return p;
            }
        }

        i = OptionMatch(start, "prs", "file");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (q == NULL) // Must have a value
                return ReportMissingValue(start, "PRS_FILE");
            else
            {
                p = GetStringValue("PRS_FILE", q, prs_file);
                return p;
            }
        }
    }

    int i = OptionMatch(start, "parent", "saved");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        parent_saved = flag;
        const char *p = start + i + 1;
        if (i < 1)
             return ReportAmbiguousOption(start, "PACKAGE, PARENT_SAVED, PARSETABLE-INTERFACES, PREFIX, PRIORITY, PROGRAMMING_LANGUAGE, PRS_FILE");
        else if (i < 2)
             return ReportAmbiguousOption(start, "PACKAGE, PARENT_SAVED, PARSETABLE-INTERFACES");
        return (ValuedOption(p) ? ReportValueNotRequired(start, "PARENT_SAVED") : p);
    }

    return ClassifyBadOption(start, flag);
}

//
// quiet
//
const char *Option::ClassifyQ(const char *start, bool flag)
{
    int i = strxsub(start + 1, "uiet");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        quiet = flag;
        const char *p = start + i + 1;
        return (ValuedOption(p) ? ReportValueNotRequired(start, "QUIET") : p);
    }

    return ClassifyBadOption(start, flag);
}

//
// read_reduce
// remap_terminals
//
const char *Option::ClassifyR(const char *start, bool flag)
{
    int i = OptionMatch(start, "read", "reduce");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        read_reduce = flag;
        const char *p = start + i + 1;
        return (i < 2 ? ReportAmbiguousOption(start, "READ_REDUCE, REMAP_TERMINALS")
                      : ValuedOption(p) ? ReportValueNotRequired(start, "READ_REDUCE") : p);
    }

    i = OptionMatch(start, "remap", "terminals");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        remap_terminals = flag;
        const char *p = start + i + 1;
        return (ValuedOption(p) ? ReportValueNotRequired(start, "REMAP_TERMINALS") : p);
    }

    return ClassifyBadOption(start, flag);
}

//
// // save_parent
// scopes
// serialize
// shift_default
// single_productions
// slr
// soft_keywords
// states
// suffix
// sym_file
//
const char *Option::ClassifyS(const char *start, bool flag)
{
    if (flag)
    {
        int i = strxsub(start + 1, "uffix");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);

            if (q == NULL)
            {
                if (i > 0)
                     return ReportMissingValue(start, "SUFFIX");
            }
            else
            {
                p = GetStringValue("SUFFIX", q, suffix);
                return p;
            }
        }

        i = OptionMatch(start, "sym", "file");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (q == NULL) // Must have a value
            {
                if (i > 0)
                     return ReportMissingValue(start, "SYM_FILE");
            }
            else
            {
                p = GetStringValue("SYM_FILE", q, sym_file);
                return p;
            }
        }
    }

    // int i = OptionMatch(start, "save", "parent");
    // if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    // {
    //     save_parent = flag;
    //     const char *p = start + i + 1;
    //     return (i < 1
    //               ? ReportAmbiguousOption(start, "SAVE_PARENT, SCOPES, SERIALIZE, SOFT_KEYWORDS, SHIFT_DEFAULT, SINGLE_PRODUCTIONS, SLR, STATES")
    //               : ValuedOption(p) ? ReportValueNotRequired(start, "SAVE_PARENT") : p);
    // }
    //
    // i = strxsub(start + 1, "copes");
    // if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    // {
    //     scopes = flag;
    //     const char *p = start + i + 1;
    //     return (ValuedOption(p) ? ReportValueNotRequired(start, "SCOPES") : p);
    // }

    int i = strxsub(start + 1, "copes");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        scopes = flag;
        const char *p = start + i + 1;
        return (i < 1 ? ReportAmbiguousOption(start, "SCOPES, SERIALIZE, SOFT_KEYWORDS, SHIFT_DEFAULT, SINGLE_PRODUCTIONS, SLR, STATES")
                      : ValuedOption(p) ? ReportValueNotRequired(start, "SCOPES") : p);
    }

    i = strxsub(start + 1, "erialize");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        serialize = flag;
        const char *p = start + i + 1;
        return (ValuedOption(p) ? ReportValueNotRequired(start, "SERIALIZE") : p);
    }

    i = OptionMatch(start, "soft", "keywords");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        soft_keywords = flag;
        const char *p = start + i + 1;
        return (ValuedOption(p) ? ReportValueNotRequired(start, "SOFT_KEYWORDS") : p);
    }

    i = OptionMatch(start, "shift", "default");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        shift_default = flag;
        const char *p = start + i + 1;
        return (ValuedOption(p) ? ReportValueNotRequired(start, "SHIFT_DEFAULT") : p);
    }

    i = OptionMatch(start, "single", "productions");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        single_productions = flag;
        const char *p = start + i + 1;
        return (ValuedOption(p) ? ReportValueNotRequired(start, "SINGLE_PRODUCTIONS") : p);
    }

    i = strxsub(start + 1, "lr");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        slr = flag;
        const char *p = start + i + 1;
        return (ValuedOption(p) ? ReportValueNotRequired(start, "SLR") : p);
    }

    i = strxsub(start + 1, "tates");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        states = flag;
        const char *p = start + i + 1;
        return (ValuedOption(p) ? ReportValueNotRequired(start, "STATES") : p);
    }

    return ClassifyBadOption(start, flag);
}

//
// tab_file
// table
// template
// trace
//
const char *Option::ClassifyT(const char *start, bool flag)
{
    if (flag)
    {
        int i = OptionMatch(start, "tab", "file");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (q == NULL)
            {
                return (i < 3 && strxsub(start, "tab") == i + 1)
                           ? ReportAmbiguousOption(start, "TABLE, TAB_FILE")
                           : ReportMissingValue(start, "TAB_FILE");
            }
            else
            {
                p = GetStringValue("TAB_FILE", q, tab_file);
                return (i == 0 ? ReportAmbiguousOption(start, "TAB_FILE, TABLE, TEMPLATE, TRACE")
                               : i < 3 && (! (i == 1 && (start[1] == 'f' || start[1] == 'F')))
                                        ? ReportAmbiguousOption(start, "TAB_FILE, TABLE")
                                        : p);
            }
        }

        i = strxsub(start + 1, "emplate");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);

            if (q == NULL)
            {
                if (i > 0)
                     return ReportMissingValue(start, "TEMPLATE");
            }
            else
            {
                p = GetStringValue("TEMPLATE", q, template_name);
                return p;
            }
        }

        i = strxsub(start + 1, "railers");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (q == NULL)
            {
                if (i > 0)
                    return ReportMissingValue(start, "TRAILERS");
            }
            else
            {
                if (*q == '(')
                {
                    const char *filename;
                    p = GetStringValue("Filename in TRAILERS", q + 1, filename);
                    if (*filename == '*')
                    {
                        char *name = new char[strlen(filename) + strlen(file_prefix)];
                        strcpy(name, file_prefix);
                        strcat(name, filename + 1);
                        filename = name;
                        temp_string.Next() = name;
                    }

                    const char *block_begin;
                    p = GetStringValue("Block Begin Marker in TRAILERS", CleanUp(p), block_begin);
                    const char *block_end;
                    p = GetStringValue("Block End Marker in TRAILERS", CleanUp(p), block_end);
                    p = CleanUp(p);

                    if (*p == ')')
                    {
                        ProcessTrailer(filename, block_begin, block_end);
                        return ++p;
                    }

                    report.Put("***ERROR: Illegal TRAILERS option specified: ");
                    while (start < p)
                         report.PutChar(*start++);
                    report.Put(". A value of the form \"(filename,begin_block_marker,end_block_marker)\" was expected.\n");
                    return_code = 12;

                    return p;
                }
            }
        }
    }

    int i = strxsub(start + 1, "able");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        const char *p = start + i + 1,
                   *q = ValuedOption(p);
        table = flag;
        if (i == 0)
             return ReportAmbiguousOption(start, "TABLE, TRACE");
        else if (q == NULL)
             return p;
        else if (flag) // Cannot assign a value to "notable"
        {
            const char *value;
            p = GetStringValue("TABLE", q, value);
            int length = strlen(value);
            if (strxsub(value, "none") == length)
            {
                 table = false;
                 programming_language = XML;
            }
            else if(strxsub(value, "xml") == length)
                 programming_language = XML;
            else if (strxsub(value, "c") == length)
                 programming_language = C;
            else if (strxsub(value, "cpp") == length || strxsub(value, "c++") == length)
                 programming_language = CPP;
            else if (strxsub(value, "java") == length)
                 programming_language = JAVA;
            else if (strxsub(value, "plx") == length)
                 programming_language = PLX;
            else if (strxsub(value, "plxasm") == length)
                 programming_language = PLXASM;
            else if (strxsub(value, "ml") == length)
                 programming_language = ML;
            else
            {
                report.Put("***ERROR: \"");
                report.Put(value);
                report.Put("\" is an invalid value for option \"");
                for (int k = 0; k < i + 1; k++)
                    report.PutChar(start[k]);
                report.Put("\"\n");
                return_code = 12;
            }

            return p;
        }
    }

    i = strxsub(start + 1, "race");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        const char *p = start + i + 1,
                   *q = ValuedOption(p);
        if (q == NULL)
        {
            trace = (flag ? CONFLICTS : NONE);
            return p;
        }
        else if (flag) // Cannot assign a value to "notrace"
        {
            const char *value;
            p = GetStringValue("TRACE", q, value);
            int length = strlen(value);
            if (strxsub(value, "none") == length)
                 trace = NONE;
            else if (strxsub(value, "conflicts") == length)
                 trace = CONFLICTS;
            else if (strxsub(value, "full") == length)
                 trace = FULL;
            else
            {
                report.Put("***ERROR: \"");
                report.Put(value);
                report.Put("\" is an invalid value for option \"");
                for (int k = 0; k < i + 1; k++)
                    report.PutChar(start[k]);
                report.Put("\"\n");
                return_code = 12;
            }

            return p;
        }
    }

    return ClassifyBadOption(start, flag);
}

//
// variables
// verbose
// visitor
// visitor_type
//
const char *Option::ClassifyV(const char *start, bool flag)
{
    int i = strxsub(start + 1, "ariables");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        const char *p = start + i + 1,
                   *q = ValuedOption(p);
        if (i < 1)
            ReportAmbiguousOption(start, "VARIABLES, VERBOSE, VISITOR, VISITOR_TYPE");

        if (q == NULL)
        {
            variables = (flag ? BOTH : NONE);
            return p;
        }
        else if (flag) // Cannot assign a value to "novariables"
        {
            const char *value;
            p = GetStringValue("VARIABLES", q, value);
            int length = strlen(value);
            if (strxsub(value, "none") == length)
                 variables = NONE;
            else if (strxsub(value, "both") == length)
                 variables = BOTH;
            else if (strxsub(value, "terminals") == length)
                 variables = TERMINALS;
            else if (strxsub(value, "nt") == length ||
                     strxsub(value, "nonterminals") == length ||
                     strxsub(value, "non_terminals") == length)
                 variables = NON_TERMINALS;
            else
            {
                report.Put("***ERROR: \"");
                report.Put(value);
                report.Put("\" is an invalid value for option \"");
                for (int k = 0; k < i + 1; k++)
                    report.PutChar(start[k]);
                report.Put("\"\n");
                return_code = 12;
            }

            return p;
        }
    }

    i = strxsub(start + 1, "erbose");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        verbose = flag;
        const char *p = start + i + 1;
        return (ValuedOption(p) ? ReportValueNotRequired(start, "VERBOSE") : p);
    }

    i = strxsub(start + 1, "isitor");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        const char *p = start + i + 1,
                   *q = ValuedOption(p);

        if (q == NULL)
        {
            visitor = (flag ? DEFAULT : NONE);
            return p;
        }
        else if (flag) // Cannot assign a value to "novisitor"
        {
            const char *value;
            p = GetStringValue("VISITOR", q, value);
            int length = strlen(value);
            if (strxsub(value, "none") == length)
                 visitor = NONE;
            else if (strxsub(value, "default") == length)
                 visitor = DEFAULT;
            else if (strxsub(value, "preorder") == length)
                 visitor = PREORDER;
            else
            {
                report.Put("***ERROR: \"");
                report.Put(value);
                report.Put("\" is an invalid value for option \"");
                for (int k = 0; k < i + 1; k++)
                    report.PutChar(start[k]);
                report.Put("\"\n");
                return_code = 12;
            }

            return p;
        }
    }

    if (flag)
    {
        i = OptionMatch(start, "visitor", "type");
        if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
        {
            const char *p = start + i + 1,
                       *q = ValuedOption(p);
            if (q == NULL)
                return ReportMissingValue(start, "VISITOR_TYPE");
            else
            {
                p = GetStringValue("VISITOR_TYPE", q, visitor_type);
                return p;
            }
        }
    }

    return ClassifyBadOption(start, flag);
}

//
// warnings
//
const char *Option::ClassifyW(const char *start, bool flag)
{
    int i = strxsub(start + 1, "arnings");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        warnings = flag;
        const char *p = start + i + 1;
        return (ValuedOption(p) ? ReportValueNotRequired(start, "WARNINGS") : p);
    }

    return ClassifyBadOption(start, flag);
}

//
// xreference
//
const char *Option::ClassifyX(const char *start, bool flag)
{
    int i = strxsub(start + 1, "reference");
    if (start[i + 1] == '=' || IsDelimiter(start[i + 1]))
    {
        xref = flag;
        const char *p = start + i + 1;
        return (ValuedOption(p) ? ReportValueNotRequired(start, "REFERENCE") : p);
    }

    return ClassifyBadOption(start, flag);
}

//
//
//
const char *Option::ClassifyBadOption(const char *start, bool flag)
{
    report.Put("***ERROR: \"");
    for (start = (flag ? start : start - 2); ! IsDelimiter(*start); start++)
        report.PutChar(*start);
    while (IsSpace(*start))
        start++;
    if (*start == '=')
    {
        report.PutChar(*start++);
        while (IsSpace(*start))
            start++;
        while (!IsDelimiter(*start))
            report.PutChar(*start++);
    }
    report.Put("\" is an invalid option\n");
    return_code = 12;

    return start;
}

//
//
//
const char *Option::ReportAmbiguousOption(const char *start, const char *msg)
{
    report.Put("***ERROR: The option \"");
    while (!IsDelimiter(*start))
        report.PutChar(*start++);
    while (IsSpace(*start))
        start++;
    if (*start == '=')
    {
        report.PutChar(*start++);
        while (IsSpace(*start))
            start++;
        while (!IsDelimiter(*start))
            report.PutChar(*start++);
    }
    report.Put("\" is ambiguous: ");
    report.Put(msg);
    report.Put(" ?\n");
    return_code = 12;

    return start;
}

const char *Option::ReportMissingValue(const char *start, const char *option)
{
    report.Put("***ERROR: A value is required for the option \"");
    report.Put(option);
    report.Put("\". The input specified was \"");
    while (!IsDelimiter(*start))
        report.PutChar(*start++);
    while (IsSpace(*start))
        start++;
    if (*start == '=')
    {
        report.PutChar(*start++);
        while (IsSpace(*start))
            start++;
        while (!IsDelimiter(*start))
            report.PutChar(*start++);
    }
    report.Put("\"\n");
    return_code = 12;

    return start;
}

//
//
//
const char *Option::ReportValueNotRequired(const char *start, const char *option)
{
    report.Put("***ERROR: A illegal value was specified for the switch option \"");
    report.Put(option);
    report.Put("\". The input specified was \"");
    while (!IsDelimiter(*start))
        report.PutChar(*start++);
    while (IsSpace(*start))
        start++;
    if (*start == '=')
    {
        report.PutChar(*start++);
        while (IsSpace(*start))
            start++;
        while (!IsDelimiter(*start))
            report.PutChar(*start++);
    }
    report.Put("\"\n");
    return_code = 12;

    return start;
}

//
//
//
bool Option::IsDelimiter(char c)
{
    return (c == NULL_CHAR || c == ',' || c == ' ');
}

//
// Clean front of string
//
const char *Option::CleanUp(const char *parm)
{
    while (*parm != NULL_CHAR && IsDelimiter(*parm))
        parm++;
    return parm;
}

//
//
//
const char *Option::ValuedOption(const char *p)
{
    while (IsSpace(*p))
        p++;
    if (*p == '=')
    {
        for (p++; IsSpace(*p); p++)
            ;
        return p;
    }

    return NULL;
}

//
//
//
const char *Option::GetValue(const char *p, const char *&value)
{
    const char *tail = p;

    while (! IsDelimiter(*tail))
        tail++;
    int length = tail - p;

    char *temp = new char[length + 1];
    strncpy(temp, p, length);
    temp[length] = NULL_CHAR;
    value = temp;
    temp_string.Next() = temp;

    return tail;
}


//
// Starting at location p in a string pick up a value (which may be
// quoted) and assign it to the reference parameter value. Return
// the ending location of the value in the string.
//
const char *Option::GetStringValue(const char *target, const char *p, const char *&value)
{
    const char *tail = p;
    char quote = (*tail == '\'' || *tail == '\"' ? *tail : NULL_CHAR);
    int length;

    //
    //
    //
    if (quote != NULL_CHAR)
    {
        for (tail++; (*tail != quote) && (*tail != NULL_CHAR); tail++)
            ;

        //
        // an isolated quote or a properly quoted string
        //
        if ((tail - p) == 1 || *tail == quote)
        {
            tail++;
            length = tail - p;
            CheckQuotedValue(target, p, length);
        }
        else
        {
            report.Put("***ERROR: The string ");
            for (const char *q = p; q < tail; q++)
                report.PutChar(*q);
            report.Put("\" was not properly terminated.\n");
            return_code = 12;

            p++;
            length = tail - p;
        }
    }
    else
    {
        while (! IsDelimiter(*tail))
            tail++;
        length = tail - p;
    }

    char *temp = new char[length + 1];
    strncpy(temp, p, length);
    temp[length] = NULL_CHAR;
    value = temp;
    temp_string.Next() = temp;

    return tail;
}


//
//
//
int Option::OptionMatch(const char *p, const char *word1, const char *word2)
{
    if ((p[0] == *word1 || p[0] == ToUpper(*word1)) &&
        (p[1] == *word2 || p[1] == ToUpper(*word2)) &&
        (p[2] == '=' || IsDelimiter(p[2])))
        return 1;

    int length1 = strlen(word1);
    int length = length1 + strlen(word2) + 1; // +1 for separator
    char *name = new char[length + 1];

    strcpy(name, word1);
    strcat(name, word2);
    int i = strxsub(p, name) - 1;

    strcpy(name, word1);
    strcat(name, "_");
    strcat(name, word2);

    i = Max(i, strxsub(p, name) - 1);

    name[length1] = '-';

    i = Max(i, strxsub(p, name) - 1);

    delete [] name;

    return i;
}

//
//
//
void Option::ProcessBlock(const char *filename, const char *block_begin, const char *block_end)
{
    int filename_length = strlen(filename),
        block_begin_length = strlen(block_begin),
        block_end_length = strlen(block_end);

    ActionFileSymbol *filename_symbol = action_blocks.FindFilename(filename, filename_length);
    if (filename_symbol != NULL)
    {
        report.Put("***ERROR: The action filename \"");
        report.Put(filename_symbol -> Name());
        report.Put("\" was previously associated with the begin block marker \"");
        report.Put(filename_symbol -> Block() -> Name());
        report.Put("\" and cannot be reassociated with the block marker \"");
        report.Put(block_begin);
        report.Put("\".\n");
        return_code = 12;
    }
    else
    {
        if (action_blocks.FindBlockname(block_begin, block_begin_length) != NULL)
        {
            report.Put("***ERROR: The action block string ");
            report.Put(block_begin);
            report.Put(" was previously used\n");
            return_code = 12;
        }
        else
        {
            filename_symbol = action_blocks.FindOrInsertFilename(filename, filename_length);
            (void) action_blocks.InsertBlock(BlockSymbol::MAIN_BLOCK,
                                             filename_symbol,
                                             block_begin,
                                             block_begin_length,
                                             block_end,
                                             block_end_length);
        }
    }

    return;
}


//
//
//
void Option::ProcessHeader(const char *filename, const char *block_begin, const char *block_end)
{
    int filename_length = strlen(filename),
        block_begin_length = strlen(block_begin),
        block_end_length = strlen(block_end);

    ActionFileSymbol *filename_symbol = action_blocks.FindFilename(filename, filename_length);
    if (filename_symbol != NULL)
    {
        if (action_blocks.FindBlockname(block_begin, block_begin_length) != NULL)
        {
            report.Put("***ERROR: The action block string ");
            report.Put(block_begin);
            report.Put(" was previously used\n");
            return_code = 12;
        }
        else
        {
            action_blocks.InsertBlock(BlockSymbol::HEADER_BLOCK,
                                      filename_symbol,
                                      block_begin,
                                      block_begin_length,
                                      block_end,
                                      block_end_length);
        }
    }
    else
    {
        report.Put("***ERROR: The action filename \"");
        report.Put(filename_symbol -> Name());
        report.Put("\" was not previously associated with an action block.\n");
        return_code = 12;
    }

    return;
}


//
//
//
void Option::ProcessTrailer(const char *filename, const char *block_begin, const char *block_end)
{
    int filename_length = strlen(filename),
        block_begin_length = strlen(block_begin),
        block_end_length = strlen(block_end);

    ActionFileSymbol *filename_symbol = action_blocks.FindFilename(filename, filename_length);
    if (filename_symbol != NULL)
    {
        if (action_blocks.FindBlockname(block_begin, block_begin_length) != NULL)
        {
             report.Put("***ERROR: The action block string ");
             report.Put(block_begin);
             report.Put(" was previously used\n");
             return_code = 12;
        }
        else
        {
            action_blocks.InsertBlock(BlockSymbol::TRAILER_BLOCK,
                                      filename_symbol,
                                      block_begin,
                                      block_begin_length,
                                      block_end,
                                      block_end_length);
        }
    }
    else
    {
        report.Put("***ERROR: The action filename \"");
        report.Put(filename_symbol -> Name());
        report.Put("\" was not previously associated with an action block.\n");
        return_code = 12;
    }

    return;
}


//
//
//
void Option::CheckBlockMarker(const char *marker, int length)
{
    if (length == 1)
    {
        if (*marker == escape)
        {
            report.Put("***ERROR: The escape symbol, \"");
            report.PutChar(escape);
            report.Put("\", cannot be used as a block marker.\n");
            return_code = 12;

            return;
        }
        if (*marker == or_marker)
        {
            report.Put("***ERROR: The OR_MARKER symbol, \"");
            report.PutChar(or_marker);
            report.Put("\", cannot be used as a block marker.\n");
            return_code = 12;

            return;
        }
    }

    if (strcmp(marker, "::=") == 0)
    {
        report.Put("***ERROR: \"::=\" cannot be used as a block marker.\n");
        return_code = 12;

        return;
    }

    if (strcmp(marker, "->") == 0)
    {
        report.Put("***ERROR: \"->\" cannot be used as a block marker.\n");
        return_code = 12;

        return;
    }

    return;
}


//
//
//
void Option::CheckQuotedValue(const char *target, const char *&value, int &length)
{
    if (length == 1)
    {
        if (*value == '\'')
        {
            report.Put("***ERROR: A single quote, \"\'\", is an invalid value for ");
            report.Put(target);
            report.PutChar('\n');
            return_code = 12;
        }
        else if (*value == '\"')
        {
            report.Put("***ERROR: A double quote (\") is an invalid value for ");
            report.Put(target);
            report.PutChar('\n');
            return_code = 12;
        }
    }
    else if (length > 1) // length is neither 0 nor 1
    {
        value++;
        length -= 2;
    }

    return;
}


//
//
//
void Option::ProcessOptions(const char *parm)
{
    for (parm = CleanUp(parm); *parm != NULL_CHAR; parm = CleanUp(parm))
    {
        bool flag = ! ((parm[0] == 'n' || parm[0] == 'N') &&
                       (parm[1] == 'o' || parm[1] == 'O'));

        if (! flag)
            parm += 2; // skip the "NO" prefix

        unsigned char c = (unsigned char) (*parm);
        parm = (this ->* classify_option[c < 128 ? c : 0])(parm, flag);
    }

    return;
}

//
//
//
void Option::ProcessUserOptions(char *line, int length)
{
    char *parm = new char[length + 1];
    strncpy(parm, line, length);
    parm[length] = NULL_CHAR;

    ProcessOptions(parm);

    delete [] parm;

    return;
}


void Option::ProcessCommandOptions()
{
    int parm_length = 0;
    for (int k = 1; k < argc - 1; k++)
        parm_length += strlen(argv[k]);
    parm_length++;

    char *parm = new char[parm_length + 1];
    *parm = NULL_CHAR;

    for (int m = 1; m < argc - 1; m++)
    {
        if (*(argv[m]) == '-')
            strcat(parm, argv[m] + 1);
        else if (*(argv[m]) == '\'' || *(argv[m]) == '\"')
        {
            int length = strlen(parm) - 1;
            if (parm[length] == ',')
                 strcpy(&(parm[length]), argv[m]);
            else
            {
                report.Put("***Warning: Stray quotation mark (");
                report.Put(argv[m]);
                report.Put(") specified in command line has been ignored.\n");
            }
        }
        else
        {
            strcat(parm, argv[m]);
            report.Put("***Warning: Option \"");
            report.Put(argv[m]);
            report.Put("\" is missing preceding '-'.\n");
        }
        strcat(parm, ",");
    }

    ProcessOptions(parm);

    delete [] parm;

    return;
}


//
//
//
void Option::ProcessPath(Tuple<const char *> &list, const char *path, const char *extra_name)
{
    int allocation_length = strlen(path) + (extra_name == NULL ? 0 : strlen(extra_name));
    char *str = new char[allocation_length + 3]; // 2 semicolons 1 for terminating \0
    temp_string.Next() = str;
    strcpy(str, path);

    int length;
    for (length = strlen(str) - 1; IsSpace(str[length]); length--)
        ; // remove trailing spaces
    if (str[length] != ';') // make sure that the path ends with a terminating semicolon
        str[++length] = ';';
    str[++length] = NULL_CHAR;
    if (extra_name != NULL)
    {
        strcat(str, extra_name);
        strcat(str, ";");
        length = strlen(str);
    }
 
    list.Reset();
    list.Next() = str;
    for (int i = 1; i < length; i++)
    {
        if (str[i] == ';')
        {
            str[i] = '\0';
            if (str[i + 1] != '\0')
                list.Next() = &(str[i + 1]);
        }
    }

    return;
}


const char *Option::GetPrefix(const char *filename)
{
    const char *forward_slash = strrchr(filename, '/'),
               *backward_slash = strrchr(filename, '\\'),
               *slash = (forward_slash > backward_slash ? forward_slash : backward_slash),
               *colon = strrchr(filename, ':'), // Windows files may use format: d:filename
               *separator = (colon > slash ? colon : slash);
    int length = (separator == NULL ? 0 : separator - filename + 1);
    char *result = new char[length + 1];
    memcpy(result, filename, length);
    result[length] = '\0';
    temp_string.Next() = result;
    return result;
}


const char *Option::GetType(const char *filename)
{
    const char *forward_slash = strrchr(filename, '/'),
               *backward_slash = strrchr(filename, '\\'),
               *slash = (forward_slash > backward_slash ? forward_slash : backward_slash),
               *colon = strrchr(filename, ':'), // Windows files may use format: d:filename
               *separator = (colon > slash ? colon : slash),
               *start = (separator ? separator + 1 : filename),
               *dot = strrchr(filename, '.');
    int length = (dot == NULL ? strlen(start) : dot - start);
    char *result = new char[length + 1];
    memcpy(result, start, length);
    result[length] = '\0';
    temp_string.Next() = result;
    return result;
}


void Option::CompleteOptionProcessing()
{
    //
    //
    //
    if (escape == ' ')
    {
        escape = (programming_language == JAVA ||
                  programming_language == C ||
                  programming_language == CPP
                             ? '$'
                             : '%');
    }

    //
    //
    //
    if (package == NULL)
    {
        char *temp = new char[1];
        *temp = NULL_CHAR;
        package = temp;
        temp_string.Next() = package;
    }

    //
    //
    //
    if (prefix == NULL)
    {
        char *temp = new char[1];
        *temp = NULL_CHAR;
        prefix = temp;
        temp_string.Next() = prefix;
    }

    //
    //
    //
    if (suffix == NULL)
    {
        char *temp = new char[1];
        *temp = NULL_CHAR;
        suffix = temp;
        temp_string.Next() = suffix;
    }

    //
    //
    //
    if (template_name == NULL)
    {
        char *temp = new char[1];
        *temp = NULL_CHAR;
        template_name = temp;
        temp_string.Next() = template_name;
    }

    assert(file_prefix);

    //
    // If no action block was specified for the default
    // action filename and the default block-begin marker,
    // we specify one here.
    //
    default_block = action_blocks.FindBlockname(default_block_begin, strlen(default_block_begin));
    if (default_block == NULL)
    {
        ProcessBlock("", default_block_begin, default_block_end);
        default_block = action_blocks.FindBlockname(default_block_begin, strlen(default_block_begin));
    }
    default_action_file = default_block -> ActionfileSymbol();
    default_action_prefix = GetPrefix(default_action_file -> Name());
    action_type = GetType(default_action_file -> Name());

    //
    // More than one action block specified? check for errors.
    //
    for (int j = 0; j < action_blocks.NumActionBlocks(); j++)
    {
        BlockSymbol *block = action_blocks[j];
        CheckBlockMarker(block -> BlockBegin(), block -> BlockBeginLength());
        CheckBlockMarker(block -> BlockEnd(), block -> BlockEndLength());
    }

    for (int c = 0; c < 256; c++)
    {
        Tuple<BlockSymbol *> &blocks = action_blocks.Blocks(c);
        for (int i = 0; i < blocks.Length(); i++)
        {
            BlockSymbol *block1 = blocks[i];

            for (int k = i + 1; k < blocks.Length(); k++)
            {
                BlockSymbol *block2 = blocks[k];
                if (block1 -> BlockBeginLength() < block2 -> BlockBeginLength())
                {
                    if (strncmp(block1 -> BlockBegin(), block2 -> BlockBegin(), block1 -> BlockBeginLength()) == 0)
                    {
                        report.Put("***ERROR: The block marker \"");
                        report.Put(block1 -> BlockBegin());
                        report.Put("\" is a substring of the block marker \"");
                        report.Put(block2 -> BlockBegin());
                        report.Put("\".\n");
                        return_code = 12;
                    }
                }
                else if (block2 -> BlockBeginLength() < block1 -> BlockBeginLength())
                {
                    if (strncmp(block2 -> BlockBegin(), block1 -> BlockBegin(), block2 -> BlockBeginLength()) == 0)
                    {
                        report.Put("***ERROR: The block marker \"");
                        report.Put(block2 -> BlockBegin());
                        report.Put("\" is a substring of the block marker \"");
                        report.Put(block1 -> BlockBegin());
                        report.Put("\".\n");
                        return_code = 12;
                    }
                }
            }
        }
    }

    //
    //
    //
    if (or_marker == escape)
    {
        report.Put("***ERROR: The escape symbol, \"");
        report.PutChar(escape);
        report.Put("\", cannot be used as the OR_MARKER.\n");
        return_code = 12;
    }
    else if (or_marker == ':')
    {
         report.Put("***ERROR: \":\" cannot be used as the OR_MARKER.\n");
         return_code = 12;
    }
    else if (or_marker == '<')
    {
         report.Put("***ERROR: \"<\" cannot be used as the OR_MARKER.\n");
         return_code = 12;
    }
    else if (or_marker == '-')
    {
         report.Put("***ERROR: \"-\" cannot be used as the OR_MARKER.\n");
         return_code = 12;
    }
    else if (or_marker == '\'')
    {
         report.Put("***ERROR: \"'\" cannot be used as the OR_MARKER.\n");
         return_code = 12;
    }
    else if (or_marker == '\"')
    {
         report.Put("***ERROR: \" cannot be used as the OR_MARKER.\n");
         return_code = 12;
    }

    //
    //
    //
    if (escape == ':')
    {
         report.Put("***ERROR: \":\" cannot be used as ESCAPE.\n");
         return_code = 12;
    }
    else if (escape == '<')
    {
         report.Put("***ERROR: \"<\" cannot be used as ESCAPE.\n");
         return_code = 12;
    }
    else if (escape == '-')
    {
         report.Put("***ERROR: \"-\" cannot be used as ESCAPE.\n");
         return_code = 12;
    }
    else if (escape == '\'')
    {
         report.Put("***ERROR: \"'\" cannot be used as ESCAPE.\n");
         return_code = 12;
    }
    else if (escape == '\"')
    {
         report.Put("***ERROR: \" cannot be used as ESCAPE.\n");
         return_code = 12;
    }

    if (ast_directory == NULL)
    {
        char *temp = new char[1];
        *temp = NULL_CHAR;
        ast_directory = temp;
        temp_string.Next() = ast_directory;
    }

    //
    //
    //
    if (automatic_ast != NONE)
    {
        if (variables == NONE)
            variables = BOTH;

        if (automatic_ast == TOPLEVEL)
        {
            assert(package);       // package should always be processed first 
            assert(ast_directory); // ast_directory should always be processed first 
            if (*ast_directory == '\0')
            {
                ast_package = package;

                char *temp = new char[1];
                *temp = NULL_CHAR;
                ast_directory = temp;
                temp_string.Next() = ast_directory;

                int num_dots = 0,
                    ast_package_length = strlen(ast_package);
                for (int i = 0; i < ast_package_length; i++)
                {
                    if (ast_package[i] == '.')
                        num_dots++;
                }
                const char *prefix = "../";
                int prefix_length = strlen(prefix) * (num_dots + 1),
                    directory_length = prefix_length + ast_package_length + 1;
                temp = new char[directory_length + 1];
                ast_directory_prefix = temp;
                temp_string.Next() = ast_directory_prefix;
                if (strlen(ast_package) == 0)
                    strcpy(temp, prefix + 1);
                else
                {
                    strcpy(temp, prefix);
                    for (int i = 0; i < num_dots; i++)
                        strcat(temp, prefix);
                    for (int i = 0, j = prefix_length; i < ast_package_length; i++, j++)
                         temp[j] = (ast_package[i] == '.' ? '/' : ast_package[i]);
                    strcpy(&temp[prefix_length + ast_package_length], "/");
                }
            }
            else
            {
                //
                // Now construct the ast_directory_prefix
                //
                int directory_length = strlen(ast_directory);
                char *temp = new char[directory_length + 2];
                ast_directory_prefix = temp;
                temp_string.Next() = ast_directory_prefix;
                strcpy(temp, ast_directory);
                if (temp[directory_length - 1] != '/' &&
                    temp[directory_length - 1] != '\\')
                {
                    strcat(temp, "/");
                    directory_length++;
                }
    
                int start_index = (Code::IsAlpha(temp[0]) && temp[1] == ':'
                                        ? 2 // possible for WIN32 only
                                        : 0);
                if (temp[start_index] == '/' || temp[start_index] == '\\')
                {
                    temp[start_index] = '/';  // make sure it's a forward slash
                    start_index++;
                }

                int i = start_index;
                while (i < directory_length)
                {
                    while (temp[i] == '.')
                        i++;
                    if (temp[i] == '/' || temp[i] == '\\')
                    {
                        temp[i] = '/';  // make sure it's a forward slash
                        i++;
                        continue;
                    }
                    break;
                }

                char *start_ast_package = &temp[i];

                while (i < directory_length)
                {
                    char *name = &temp[i];
                    while (temp[i] != '/' && temp[i] != '\\')
                        i++;
                    temp[i] = '\0'; // test this prefix
                    if (! Code::IsValidVariableName(name))
                    {
                        report.Put("***ERROR: The subpackage \"");
                        report.Put(name);
                        report.Put("\" specified in the ast_directory option is not a valid Java variable name");
                        return_code = 12;

                        break;
                    }
                    int rc = SystemMkdir(temp);
                    if (rc == ENOENT)
                    {
                        report.Put("***ERROR: Unable to create ast directory: ");
                        report.Put(temp);
                        return_code = 12;
    
                        break;
                    }
                    else if (rc == ENOTDIR)
                    {
                        report.Put("***ERROR: The file: ");
                        report.Put(temp);
                        report.Put(", is not a directory");
                        return_code = 12;

                        break;
                    }
                    temp[i] = '/';  // restore slash
                    i++;
                }

                //
                // Remove all leading "./"
                //
                for (i = 0; temp[i] == '.' && temp[i + 1] == '/'; i += 2)
                    ;

                //
                // Check whether or not the ast package is a subpackage of the
                // of the main package. If so, make sure that the main package
                // is not the unnamed package.
                //
                char *temp_ast_package;
                if (start_ast_package == &temp[i])
                {
                    if (*package == NULL_CHAR)
                    {
                        temp_ast_package = new char[1];
                        *temp_ast_package = NULL_CHAR;

                        report.Put("***ERROR: The ast package cannot be a subpackage of the unnamed package."
                                   " Please specify a package name using the package option");
                        return_code = 12;
                    }
                    else
                    {
                        int package_length = strlen(package),
                            ast_package_length = strlen(start_ast_package) - 1,
                            length = package_length + 1 + ast_package_length;
                        temp_ast_package = new char[length + 1];
                        strcpy(temp_ast_package, package);
                        strcat(temp_ast_package, ".");
                        strncat(temp_ast_package, start_ast_package, ast_package_length);
                        temp_ast_package[length] = '\0';
                    }
                }
                else
                {
                    int length = strlen(start_ast_package) - 1;
                    temp_ast_package = new char[length + 1];
                    strncpy(temp_ast_package, start_ast_package, length);
                    temp_ast_package[length] = '\0';
                }

                //
                // Replace all slashes (/) by dots (.) in package name.
                //
                for (char *s = temp_ast_package; *s != '\0'; s++)
                {
                    if (s[i] == '/')
                        s[i] = '.';
                }

                ast_package = temp_ast_package;
                temp_string.Next() = ast_package;
            }
        }
    }

    if (ast_package == NULL)
    {
        char *temp = new char[1];
        *temp = NULL_CHAR;
        ast_package = temp;
        temp_string.Next() = ast_package;
    }

    if (ast_type == NULL)
    {
        char *temp = new char[4];
        strcpy(temp, "Ast");
        ast_type = temp;
        temp_string.Next() = ast_type;
    }

    //
    //
    //
    if (visitor_type == NULL)
    {
        char *temp = new char[8];
        strcpy(temp, "Visitor");
        visitor_type = temp;
        temp_string.Next() = visitor_type;
    }

    //
    //
    //
    if (prs_file == NULL)
    {
        const char *file_suffix = "prs.",
                   *file_type = (programming_language == JAVA
                                     ? "java"
                                     : (programming_language == ML
                                            ? "ml"
                                            : (programming_language == PLX || programming_language == PLXASM
                                                   ? "copy"
                                                   : (programming_language == C || programming_language == CPP
                                                          ? "h"
                                                          : "xml"))));

        int filename_length = strlen(file_prefix) + strlen(file_suffix) + strlen(file_type);
        char *temp = new char[filename_length + 1];
        prs_file = temp;
        temp_string.Next() = prs_file;
        strcpy(temp, file_prefix);
        strcat(temp, file_suffix);
        strcat(temp, file_type);
    }
    prs_type = GetType(prs_file);

    //
    //
    //
    if (sym_file == NULL)
    {
        const char *file_suffix = "sym.",
                   *file_type = (programming_language == JAVA
                                     ? "java"
                                     : (programming_language == ML
                                            ? "ml"
                                            : (programming_language == PLX || programming_language == PLXASM
                                                   ? "copy"
                                                   : (programming_language == C || programming_language == CPP
                                                          ? "h"
                                                          : "xml"))));

        int filename_length = strlen(file_prefix) + strlen(file_suffix) + strlen(file_type);
        char *temp = new char[filename_length + 1];
        sym_file = temp;
        temp_string.Next() = sym_file;
        strcpy(temp, file_prefix);
        strcat(temp, file_suffix);
        strcat(temp, file_type);
    }
    sym_type = GetType(sym_file);

    //
    //
    //
    if (dat_directory == NULL)
    {
        char *temp = new char[3];
        strcpy(temp, "./");
        dat_directory = temp;
        temp_string.Next() = dat_directory;
    }

    if (dat_file == NULL)
    {
        const char *file_suffix = "dcl.",
                   *file_type = "data";

        int filename_length = strlen(file_prefix) + strlen(file_suffix) + strlen(file_type);
        char *temp = new char[filename_length + 1];
        dat_file = temp;
        temp_string.Next() = dat_file;
        strcpy(temp, file_prefix);
        strcat(temp, file_suffix);
        strcat(temp, file_type);
    }

    //
    //
    //
    if (dcl_file == NULL)
    {
        const char *file_suffix = ((programming_language == C || programming_language == CPP) ? "prs." : "dcl."),
                   *file_type = (programming_language == JAVA
                                     ? "java"
                                     : (programming_language == ML
                                            ? "ml"
                                            : (programming_language == CPP
                                                   ? "cpp"
                                                   : (programming_language == PLX
                                                          ? "copy"
                                                          : (programming_language == PLXASM
                                                                 ? "assemble"
                                                                 : (programming_language == C
                                                                        ? "c"
                                                                        : "xml"))))));

        int filename_length = strlen(file_prefix) + strlen(file_suffix) + strlen(file_type);
        char *temp = new char[filename_length + 1];
        dcl_file = temp;
        temp_string.Next() = dcl_file;
        strcpy(temp, file_prefix);
        strcat(temp, file_suffix);
        strcat(temp, file_type);
    }

    //
    //
    //
    if (def_file == NULL)
    {
        const char *file_suffix = "def.",
                   *file_type = (programming_language == JAVA
                                     ? "java"
                                     : (programming_language == ML
                                            ? "ml"
                                            : (programming_language == CPP
                                                   ? "cpp"
                                                   : (programming_language == PLX || programming_language == PLXASM
                                                          ? "copy"
                                                          : (programming_language == C
                                                                 ? "c"
                                                                 : "xml")))));

        int filename_length = strlen(file_prefix) + strlen(file_suffix) + strlen(file_type);
        char *temp = new char[filename_length + 1];
        def_file = temp;
        temp_string.Next() = def_file;
        strcpy(temp, file_prefix);
        strcat(temp, file_suffix);
        strcat(temp, file_type);
    }

    //
    // Check export_terminals option.
    //
    if (exp_file == NULL)
    {
        const char *file_suffix = "exp.",
                   *file_type = (programming_language == JAVA
                                     ? "java"
                                     : (programming_language == ML
                                            ? "ml"
                                            : (programming_language == PLX || programming_language == PLXASM
                                                  ? "copy"
                                                  : (programming_language == C || programming_language == CPP
                                                         ? "h"
                                                         : "xml"))));

        int filename_length = strlen(file_prefix) + strlen(file_suffix) + strlen(file_type);
        char *temp = new char[filename_length + 1];
        exp_file = temp;
        temp_string.Next() = exp_file;
        strcpy(temp, file_prefix);
        strcat(temp, file_suffix);
        strcat(temp, file_type);
    }
    exp_type = GetType(exp_file);

    if (exp_prefix == NULL)
    {
        char *temp = new char[1];
        *temp = NULL_CHAR;
        exp_prefix = temp;
        temp_string.Next() = exp_prefix;
    }

    if (exp_suffix == NULL)
    {
        char *temp = new char[1];
        *temp = NULL_CHAR;
        exp_suffix = temp;
        temp_string.Next() = exp_suffix;
    }

    if (factory == NULL)
    {
        char *temp = new char[5];
        strcpy(temp, "new ");
        factory = temp;
        temp_string.Next() = factory;
    }

    //
    //
    //
    if (imp_file == NULL)
    {
        const char *file_suffix = "imp",
                   *file_type = (programming_language == XML
                                     ? ".xml"
                                     : (programming_language == C || programming_language == CPP
                                            ? ".h"
                                            : (programming_language == PLX || programming_language == PLXASM
                                                   ? ".copy"
                                                   : (programming_language == JAVA
                                                          ? ".java"
                                                          : ".ml"))));
        int filename_length = strlen(file_prefix) + strlen(file_suffix) + strlen(file_type);
        char *temp = new char[filename_length + 1];
        imp_file = temp;
        temp_string.Next() = imp_file;
        strcpy(temp, file_prefix);
        strcat(temp, file_suffix);
        strcat(temp, file_type);
    }

    //
    //
    //
    if (soft_keywords)
    {
        lalr_level = 1;
        single_productions = false;
        backtrack = true;
    }

    //
    //
    //
    if (verbose)
    {
        first = true;
        follow = true;
        list = true;
        states = true;
        xref = true;
        warnings = true;
    }

    return;
}


//
// Here we print all options set by the user.
//
void Option::PrintOptionsInEffect()
{
    if (this -> quiet)
        return;

    assert(grm_file);
    report.Put("\nOptions in effect for ");
    report.Put(grm_file);
    report.Put(":\n\n");
    for (int i = 0; i < action_blocks.NumActionBlocks(); i++)
    {
        BlockSymbol *block = action_blocks[i];

        report.Put("    ");
        report.Put(block -> Kind() == BlockSymbol::MAIN_BLOCK
                                    ? "ACTION-BLOCK"
                                    : block -> Kind() == BlockSymbol::HEADER_BLOCK
                                             ? "HEADERS"
                                             : "TRAILERS");
        report.Put("=(\"");
        report.Put(block -> ActionfileSymbol() -> Name());
        report.Put("\",\"");
        report.Put(block -> BlockBegin());
        report.Put("\",\"");
        report.Put(block -> BlockEnd());
        report.Put("\")");
        report.PutChar('\n');
    }
    report.PutChar('\n');

    Tuple<const char *> opt_string;

    opt_string.Next() = AllocateString("AST-DIRECTORY", ast_directory);
    opt_string.Next() = AllocateString("AST-TYPE", ast_type);
    opt_string.Next() = AllocateString(attributes ? "ATTRIBUTES" : "NOATTRIBUTES");
    opt_string.Next() = AllocateString(automatic_ast == NONE
                                           ? "NOAUTOMATIC-AST"
                                           : automatic_ast == NESTED
                                                 ? "AUTOMATIC-AST=NESTED"
                                                 : "AUTOMATIC-AST=TOPLEVEL");
    opt_string.Next() = AllocateString(backtrack ? "BACKTRACK" : "NOBACKTRACK");
    if (byte)
        opt_string.Next() = AllocateString("BYTE");
    opt_string.Next() = AllocateString(conflicts ? "CONFLICTS" : "NOCONFLICTS");
    opt_string.Next() = AllocateString("DAT-DIRECTORY", dat_directory);
    opt_string.Next() = AllocateString("DAT-FILE", dat_file);
    opt_string.Next() = AllocateString("DCL-FILE", dcl_file);
    opt_string.Next() = AllocateString(debug ? "DEBUG" : "NODEBUG");
    opt_string.Next() = AllocateString("DEF-FILE", def_file);
    opt_string.Next() = AllocateString(edit ? "EDIT" : "NOEDIT");
    opt_string.Next() = AllocateString(error_maps ? "ERROR-MAPS" : "NOERROR-MAPS");
    opt_string.Next() = AllocateString("ESCAPE", escape);

    //
    // Reconstruct the export terminals information
    //
    {
        char *export_info = new char[strlen(exp_file) +
                                     strlen(exp_prefix) +
                                     strlen(exp_suffix) + 11]; // 11 = 6 quotes, 2 commas, 2 parentheses, \0
        strcpy(export_info, "(\"");
        strcat(export_info, exp_file);
        strcat(export_info, "\",\"");
        strcat(export_info, exp_prefix);
        strcat(export_info, "\",\"");
        strcat(export_info, exp_suffix);
        strcat(export_info, "\")");
        opt_string.Next() = AllocateString("EXPORT-TERMINALS", export_info);
        delete [] export_info;
    }
    if (extends_parsetable == NULL)
         opt_string.Next() = AllocateString("EXTENDS-PARSETABLE");
    else if (strlen(extends_parsetable) == 0)
         opt_string.Next() = AllocateString("EXTENDS-PARSETABLE");
    else opt_string.Next() = AllocateString("EXTENDS-PARSETABLE", extends_parsetable);
    opt_string.Next() = AllocateString("FACTORY", factory);
    opt_string.Next() = AllocateString("FILE-PREFIX", file_prefix);
    if (filter)
        opt_string.Next() = AllocateString("FILTER", filter);
    opt_string.Next() = AllocateString(first ? "FIRST" : "NOFIRST");
    opt_string.Next() = AllocateString(follow ? "FOLLOW" : "NOFOLLOW");
    opt_string.Next() = AllocateString(goto_default ? "GOTO-DEFAULT" : "NOGOTO-DEFAULT");
    opt_string.Next() = AllocateString("GRM-FILE", grm_file);
    if (imp_file)
        opt_string.Next() = AllocateString("IMP-FILE", imp_file);
    if(import_terminals)
        opt_string.Next() = AllocateString("IMPORT-TERMINALS", import_terminals);
    opt_string.Next() = AllocateString("INCLUDE-DIRECTORY", include_directory);
    if (! slr)
        opt_string.Next() = AllocateString("LALR", lalr_level);
    opt_string.Next() = AllocateString(list ? "LIST" : "NOLIST");
    opt_string.Next() = AllocateString("MARGIN", margin);
    opt_string.Next() = AllocateString("MAX-CASES", max_cases);
    opt_string.Next() = AllocateString(names == MAXIMUM
                                              ? "NAMES=MAXIMUM"
                                              : (names == MINIMUM
                                                        ? "NAMES=MINIMUM"
                                                        : "NAMES=OPTIMIZED"));
    opt_string.Next() = AllocateString(nt_check ? "NT-CHECK" : "NONT-CHECK");
    opt_string.Next() = AllocateString("OR_MARKER", or_marker);
    opt_string.Next() = AllocateString("PACKAGE", package);
    opt_string.Next() = AllocateString(parent_saved ? "PARENT-SAVED" : "NOPARENT-SAVE");
    if (parsetable_interfaces == NULL)
         opt_string.Next() = AllocateString("NOPARSETABLE-INTERFACES");
    else if (strlen(parsetable_interfaces) == 0)
         opt_string.Next() = AllocateString("PARSETABLE-INTERFACES");
    else opt_string.Next() = AllocateString("PARSETABLE-INTERFACES", parsetable_interfaces);
    opt_string.Next() = AllocateString("PREFIX", prefix);
    opt_string.Next() = AllocateString(priority ? "PRIORITY" : "NOPRIORITY");
    opt_string.Next() = AllocateString(programming_language == C
                                           ? "PROGRAMMING_LANGUAGE=C"
                                           : (programming_language == CPP
                                                  ? "PROGRAMMING_LANGUAGE=CPP"
                                                  : (programming_language == JAVA
                                                         ? "PROGRAMMING_LANGUAGE=JAVA"
                                                         : (programming_language == PLX
                                                                ? "PROGRAMMING_LANGUAGE=PLX"
                                                                : (programming_language == PLXASM
                                                                        ? "PROGRAMMING_LANGUAGE=PLXASM"
                                                                        : (programming_language == ML
                                                                               ? "PROGRAMMING_LANGUAGE=ML"
                                                                               : "PROGRAMMING_LANGUAGE=XML"))))));
    opt_string.Next() = AllocateString("PRS-FILE", prs_file);
    opt_string.Next() = AllocateString(quiet ? "QUIET" : "NOQUIET");
    opt_string.Next() = AllocateString(read_reduce ? "READ-REDUCE" : "NOREAD-REDUCE");
    opt_string.Next() = AllocateString(remap_terminals ? "REMAP-TERMINALS" : "NOREMAP-TERMINALS");
    opt_string.Next() = AllocateString(scopes ? "SCOPES" : "NOSCOPES");
    opt_string.Next() = AllocateString(serialize ? "SERIALIZE" : "NOSERIALIZE");
    opt_string.Next() = AllocateString(shift_default ? "SHIFT-DEFAULT" : "NOSHIFT-DEFAULT");
    if (! byte)
        opt_string.Next() = AllocateString("SHORT");
    opt_string.Next() = AllocateString(single_productions ? "SINGLE-PRODUCTIONS" : "NOSINGLE-PRODUCTIONS");
    if (slr)
        opt_string.Next() = AllocateString("SLR");
    opt_string.Next() = AllocateString(soft_keywords ? "SOFT-KEYWORDS" : "NOSOFT-KEYWORDS");
    opt_string.Next() = AllocateString(states ? "STATES" : "NOSTATES");
    opt_string.Next() = AllocateString("SUFFIX", suffix);
    opt_string.Next() = AllocateString("SYM-FILE", sym_file);
    opt_string.Next() = AllocateString("TAB-FILE", tab_file);
    opt_string.Next() = AllocateString(table ? "TABLE" : "NOTABLE");
    opt_string.Next() = AllocateString("TEMPLATE", template_name);
    opt_string.Next() = AllocateString(trace == NONE
                                              ? "NOTRACE"
                                              : (trace == CONFLICTS
                                                        ? "TRACE=CONFLICTS"
                                                        : "TRACE=FULL"));
    opt_string.Next() = AllocateString(variables == NONE
                                           ? "NOVARIABLES"
                                           : (variables == BOTH
                                                  ? "VARIABLES"
                                                  : (variables == TERMINALS
                                                         ? "VARIABLES=TERMINALS"
                                                          : "VARIABLES=NONTERMINALS")));
    opt_string.Next() = AllocateString(verbose ? "VERBOSE" : "NOVERBOSE");
    opt_string.Next() = AllocateString(visitor == NONE
                                           ? "NOVISITOR"
                                           : visitor == DEFAULT
                                                 ? "VISITOR=DEFAULT"
                                                 : "VISITOR=PREORDER");
    opt_string.Next() = AllocateString("VISITOR-TYPE", visitor_type);
    opt_string.Next() = AllocateString(warnings ? "WARNINGS" : "NOWARNINGS");
    opt_string.Next() = AllocateString(xref ? "XREF" : "NOXREF");

    char output_line[81];
    strcpy(output_line, "    ");
    for (int k = 0; k < opt_string.Length(); k++)
    {
        if (strlen(opt_string[k]) > 80)
        {
            report.Put(output_line);
            report.PutChar('\n');
            report.Put("    ");
            report.Put(opt_string[k]);
            report.PutChar('\n');
        }
        else
        {
            if (strlen(output_line) + strlen(opt_string[k]) > 80)
            {
                report.Put(output_line);
                report.PutChar('\n');
                strcpy(output_line, "    ");
            }
            strcat(output_line, opt_string[k]);
            if (strlen(output_line) + 2 < 80)
                strcat(output_line, "  ");
        }

        delete [] opt_string[k];
    }

    report.Put(output_line);
    report.Put("\n\n");

    return;
}


//
//
//
void Option::PrintOptionsList(void)
{
    cout << "\n"
         << Control::HEADER_INFO
         <<    "\n(C) Copyright IBM Corp. 1984, 2006.\n"
               "Usage: lpg [options] [filename[.extension]]\n\n"
               "Options:\n"
               "========\n\n"

               "-action=(string,string,string)                        " "\n"
               "-ast_directory[=string]                               " "\n"
               "-ast_type[=string]                                    " "\n"
               "-attributes                                           " "\n"
               "-automatic_ast[=<none|nested|toplevel>]               " "\n"
               "-backtrack                                            " "\n"
               "-byte                                                 " "\n"
               "-conflicts                                            " "\n"
               "-dat-directory=string                                 " "\n"
               "-dat-file=string                                      " "\n"
               "-dcl-file=string                                      " "\n"
               "-debug                                                " "\n"
               "-def-file=string                                      " "\n"
               "-edit                                                 " "\n"
               "-error-maps                                           " "\n"
               "-escape=character                                     " "\n"
               "-extends-parsetable=string                            " "\n"
               "-export-terminals=string                              " "\n"
               "-factory=string                                       " "\n"
               "-file-prefix=string                                   " "\n"
               "-filter=string                                        " "\n"
               "-first                                                " "\n"
               "-follow                                               " "\n"
               "-goto-default                                         " "\n"
               "-grm-file=string                                      " "\n"
               "-imp-file=string                                      " "\n"
               "-import-terminals=string                              " "\n"
               "-include-directory=string                             " "\n"
               "-lalr-level[=integer]                                 " "\n"
               "-list                                                 " "\n"
               "-margin=integer                                       " "\n"
               "-max_cases=integer                                    " "\n"
               "-names=<optimized|maximum|minimum>                    " "\n"
               "-nt-check                                             " "\n"
               "-or-marker=character                                  " "\n"
               "-package=string                                       " "\n"
               "-parent_saved                                         " "\n"
               "-parsetable-interfaces=string                         " "\n"
               "-prefix=string                                        " "\n"
               "-priority                                             " "\n"
               "-programming_language[=<xml|c|cpp|java|plx|plxasm|ml>]" "\n"
               "-prs-file=string                                      " "\n"
               "-quiet                                                " "\n"
               "-read-reduce                                          " "\n"
               "-remap-terminals                                      " "\n"
               "-scopes                                               " "\n"
               "-serialize                                            " "\n"
               "-shift-default                                        " "\n"
               "-single-productions                                   " "\n"
               "-slr                                                  " "\n"
               "-soft-keywords                                        " "\n"
               "-states                                               " "\n"
               "-suffix=string                                        " "\n"
               "-sym-file=string                                      " "\n"
               "-tab-file=string                                      " "\n"
               "-table                                                " "\n"
               "-template=string                                      " "\n"
               "-trace[=<conflicts|full>]                             " "\n"
               "-variables[=<none|both|terminals|nt|nonterminals>]    " "\n"
               "-verbose                                              " "\n"
               "-visitor[=<none|default|preorder>]                    " "\n"
               "-visitor-type[=string]                                " "\n"
               "-warnings                                             " "\n"
               "-xref                                                 "
               "\n\n"

    "Options must be separated by a space.  "
    "Any non-ambiguous initial prefix of a\n"
    "valid option may be used as an abbreviation "
    "for that option.  When an option is\n"
    "composed of two separate words, an "
    "abbreviation may be formed by concatenating\n"
    "the first character of each word.  "
    "Options that are switches may be negated by\n"
    "prefixing them with the string \"no\".  "
    "Default input file extension is \".g\"\n"

    "\nVersion " << Control::VERSION << " \n";
    /*"\nAddress comments and questions to charles@watson.ibm.com.\n";*/

    return;
}
