#ifndef option_INCLUDED
#define option_INCLUDED

#include "util.h"
#include "code.h"
#include "tuple.h"
#include "action.h"
#include "symbol.h"

class VariableSymbol;
class Option : public Code, public Util
{
public:

    enum
    {
        //
        // Possible values for option "names"
        //
        MINIMUM = 1,
        MAXIMUM = 2,
        OPTIMIZED = 3,

        //
        // Possible values for option "programming_language"
        //
        NONE = 0,
        XML = 1,
        C = 2,
        CPP = 3,
        JAVA = 4,
        PLX = 5,
        PLXASM = 6,
        ML = 7,

        //
        // Possible values for option "trace"
        //
        // NONE = 0,
        //
        CONFLICTS = 1,
        FULL = 2,

        //
        // Possible values for option "automatic_ast"
        //
        // NONE = 0,
        //
        NESTED = 1,
        TOPLEVEL = 2,

        //
        // Possible values for option "variables"
        //
        // NONE = 0,
        //
        BOTH = 1,
        NON_TERMINALS = 2,
        TERMINALS = 3,

        //
        // Possible values for option "visitor"
        //
        // NONE = 0,
        //
        DEFAULT = 1,
        PREORDER = 2
    };

    int return_code;

    Tuple<const char *> include_search_directory,
                        template_search_directory,
                        filter_file,
                        import_file;

    const char *template_directory,
               *ast_directory_prefix;

    bool attributes,
         backtrack,
         list,
         slr,
         verbose,
         first,
         follow,
         priority,
         edit,
         states,
         xref,
         nt_check,
         conflicts,
         read_reduce,
         remap_terminals,
         goto_default,
         shift_default,
         byte,
         warnings,
         single_productions,
         error_maps,
         debug,
         parent_saved,
         scopes,
         serialize,
         soft_keywords,
         table;

    int lalr_level,
        margin,
        max_cases,
        names,
        trace,
        programming_language,
        automatic_ast,
        variables,
        visitor;

    char escape,
         or_marker;

    const char *factory,
               *file_prefix,

               *grm_file,
               *lis_file,
               *tab_file,

               *dat_directory,
               *dat_file,
               *dcl_file,
               *def_file,
               *prs_file,
               *sym_file,
               *imp_file,
               *exp_file,
               *exp_prefix,
               *exp_suffix,

               *ast_directory,
               *ast_package,

               *ast_type,
               *exp_type,
               *prs_type,
               *sym_type,
               *action_type,
               *visitor_type,

               *filter,
               *import_terminals,
               *include_directory,
               *template_name,
               *extends_parsetable,
               *parsetable_interfaces,
               *package,
               *prefix,
               *suffix;

    bool quiet;

    TextBuffer report;

    Option(int argc_, const char **argv_) : argc(argc_),
                                            argv(argv_)
    {
        return_code = 0;
        quiet = false;
        automatic_ast = NONE;
        attributes = false;
        backtrack = false;
        list = false;
        slr = false;
        verbose = false;
        first = false;
        follow = false;
        priority = true;
        edit = false;
        states = false;
        xref = false;
        nt_check = false;
        conflicts = true;
        read_reduce = true;
        remap_terminals = true;
        goto_default = false;
        shift_default = false;
        byte = true;
        warnings = true;
        single_productions = false;
        error_maps = false;
        debug = false;
        parent_saved = false;
        scopes = false;
        serialize = false;
        soft_keywords = false;
        table = false;
        variables = NONE;
        visitor = NONE;
        lalr_level = 1;
        margin = 0;
        max_cases = 1024;
        names = OPTIMIZED;
        trace = CONFLICTS;
        programming_language = XML;
        escape = ' ';
        or_marker = '|';
        factory = NULL;
        file_prefix = NULL;
        grm_file = NULL;
        lis_file = NULL;
        tab_file = NULL;
        dat_directory = NULL;
        dat_file = NULL;
        dcl_file = NULL;
        def_file = NULL;
        prs_file = NULL;
        sym_file = NULL;
        imp_file = NULL;
        exp_file = NULL;
        exp_prefix = NULL;
        exp_suffix = NULL;
        ast_directory = NULL;
        ast_package = NULL;
        ast_type = NULL;
        exp_type = NULL;
        prs_type = NULL;
        sym_type = NULL;
        action_type = NULL;
        visitor_type = NULL;
        filter = NULL;
        import_terminals = NULL;
        include_directory = NULL;
        template_name = NULL;
        extends_parsetable = NULL;
        parsetable_interfaces = NULL;
        package = NULL;
        prefix = NULL;
        suffix = NULL;
        default_action_prefix = NULL;
        default_action_file = NULL;

        for (int c = 0; c < 128; c++)
            classify_option[c] = &Option::ClassifyBadOption;

        classify_option[(int) 'a'] = &Option::ClassifyA;
        classify_option[(int) 'A'] = &Option::ClassifyA;

        classify_option[(int) 'b'] = &Option::ClassifyB;
        classify_option[(int) 'B'] = &Option::ClassifyB;

        classify_option[(int) 'c'] = &Option::ClassifyC;
        classify_option[(int) 'C'] = &Option::ClassifyC;

        classify_option[(int) 'd'] = &Option::ClassifyD;
        classify_option[(int) 'D'] = &Option::ClassifyD;

        classify_option[(int) 'e'] = &Option::ClassifyE;
        classify_option[(int) 'E'] = &Option::ClassifyE;

        classify_option[(int) 'f'] = &Option::ClassifyF;
        classify_option[(int) 'F'] = &Option::ClassifyF;

        classify_option[(int) 'g'] = &Option::ClassifyG;
        classify_option[(int) 'G'] = &Option::ClassifyG;

        classify_option[(int) 'h'] = &Option::ClassifyH;
        classify_option[(int) 'H'] = &Option::ClassifyH;

        classify_option[(int) 'i'] = &Option::ClassifyI;
        classify_option[(int) 'I'] = &Option::ClassifyI;

        classify_option[(int) 'l'] = &Option::ClassifyL;
        classify_option[(int) 'L'] = &Option::ClassifyL;

        classify_option[(int) 'm'] = &Option::ClassifyM;
        classify_option[(int) 'M'] = &Option::ClassifyM;

        classify_option[(int) 'n'] = &Option::ClassifyN;
        classify_option[(int) 'N'] = &Option::ClassifyN;

        classify_option[(int) 'o'] = &Option::ClassifyO;
        classify_option[(int) 'O'] = &Option::ClassifyO;

        classify_option[(int) 'p'] = &Option::ClassifyP;
        classify_option[(int) 'P'] = &Option::ClassifyP;

        classify_option[(int) 'q'] = &Option::ClassifyQ;
        classify_option[(int) 'Q'] = &Option::ClassifyQ;

        classify_option[(int) 'r'] = &Option::ClassifyR;
        classify_option[(int) 'R'] = &Option::ClassifyR;

        classify_option[(int) 's'] = &Option::ClassifyS;
        classify_option[(int) 'S'] = &Option::ClassifyS;

        classify_option[(int) 't'] = &Option::ClassifyT;
        classify_option[(int) 'T'] = &Option::ClassifyT;

        classify_option[(int) 'v'] = &Option::ClassifyV;
        classify_option[(int) 'V'] = &Option::ClassifyV;

        classify_option[(int) 'w'] = &Option::ClassifyW;
        classify_option[(int) 'W'] = &Option::ClassifyW;

        classify_option[(int) 'x'] = &Option::ClassifyX;
        classify_option[(int) 'X'] = &Option::ClassifyX;

        //
        //
        //
        if (argc > 1)
        {
            //
            // Process the LPG_TEMPLATE and LPG_INCLUDE environment variables, if they
            // were specified.
            //
            const char *main_input_file = argv[argc - 1],
                       *grm_prefix = GetPrefix(main_input_file),
                       *lpg_template = getenv("LPG_TEMPLATE"),
                       *lpg_include = getenv("LPG_INCLUDE");

            char *temp = new char[strlen(grm_prefix) +
                                  (lpg_template == NULL ? 0 : strlen(lpg_template)) +
                                  4];
            template_directory = temp;
            temp_string.Next() = template_directory;
            strcpy(temp, (*grm_prefix == '\0' ? "./" : grm_prefix));
            if (lpg_template != NULL)
            {
                strcat(temp, ";");
                strcat(temp, lpg_template);
            }
            ProcessPath(template_search_directory, template_directory, "");

            temp = new char[strlen(grm_prefix) +
                            (lpg_include == NULL ? 0 : strlen(lpg_include)) +
                            4];
            include_directory = temp;
            temp_string.Next() = include_directory;
            strcpy(temp, (*grm_prefix == '\0' ? "./" : grm_prefix));
            if (lpg_include != NULL)
            {
                strcat(temp, ";");
                strcat(temp, lpg_include);
            }
            ProcessPath(include_search_directory, temp, "");

            int length = strlen(main_input_file);
            char *temp_file_prefix = new char[length + 3];
            file_prefix = temp_file_prefix;
            temp_string.Next() = file_prefix;

            char *temp_grm_file = new char[length + 3];
            grm_file = temp_grm_file;
            temp_string.Next() = grm_file;

            char *temp_lis_file = new char[length + 3];
            lis_file = temp_lis_file;
            temp_string.Next() = lis_file;

            char *temp_tab_file = new char[length + 3];
            tab_file = temp_tab_file;
            temp_string.Next() = tab_file;

            strcpy(temp_grm_file, main_input_file);

            int slash_index,
                dot_index = -1;
            for (slash_index = length - 1;
                 slash_index >= 0 && grm_file[slash_index] != '\\' && grm_file[slash_index] != '/';
                 slash_index--)
            {
                if (grm_file[slash_index] == '.')
                    dot_index = slash_index;
            }

            const char *slash = (slash_index >= 0 ? &grm_file[slash_index] : NULL),
                       *dot = (dot_index >= 0 ? &grm_file[dot_index] : NULL),
                       *start = (slash ? slash + 1 : grm_file);
            if (dot == NULL) // if filename has no extension, copy it.
            {
                strcpy(temp_file_prefix, start);
                strcpy(temp_lis_file, start);
                strcpy(temp_tab_file, start);

                strcat(temp_grm_file, ".g"); // add .g extension for input file
            }
            else // if file name contains an extension copy up to the dot
            {
                memcpy(temp_file_prefix, start, dot - start);
                memcpy(temp_lis_file, start, dot - start);
                memcpy(temp_tab_file, start, dot - start);
                temp_lis_file[dot - start] = '\0';
                temp_tab_file[dot - start] = '\0';
                temp_file_prefix[dot - start] = '\0';
            }

            strcat(temp_lis_file, ".l"); // add .l extension for listing file
            strcat(temp_tab_file, ".t"); // add .t extension for table file
        }

        return;
    }

    ~Option()
    {
        //
        // Release all temporary strings now to save space.
        //
        for (int i = 0; i < temp_string.Length(); i++)
            delete [] temp_string[i];

        return;
    }

    void ProcessUserOptions(char *, int);
    void ProcessCommandOptions();
    void CompleteOptionProcessing();
    void PrintOptionsInEffect();

    static void PrintOptionsList(void);

    Action &ActionBlocks() { return action_blocks; }
    BlockSymbol *DefaultBlock(void) { return default_block; }
    ActionFileSymbol *DefaultActionFile(void) { return default_action_file; }
    const char *DefaultActionPrefix(void) { return default_action_prefix; }

private:

    int argc;
    const char **argv;

    Action action_blocks;

    static const char *default_block_begin;
    static const char *default_block_end;
    BlockSymbol *default_block;
    ActionFileSymbol *default_action_file;
    const char *default_action_prefix;

    Tuple<const char *> temp_string;

    const char *AllocateString(const char *);
    const char *AllocateString(const char *, char);
    const char *AllocateString(const char *, const char *);
    const char *AllocateString(const char *, int);
    bool IsDelimiter(char);
    const char *CleanUp(const char *);
    const char *ValuedOption(const char *);
    const char *GetValue(const char *, const char *&);
    const char *GetStringValue(const char *, const char *, const char *&);
    int OptionMatch(const char *, const char *, const char *);

    const char *(Option::*classify_option[128])(const char *, bool);

    const char *ClassifyA(const char *, bool = true);
    const char *ClassifyB(const char *, bool = true);
    const char *ClassifyC(const char *, bool = true);
    const char *ClassifyD(const char *, bool = true);
    const char *ClassifyE(const char *, bool = true);
    const char *ClassifyF(const char *, bool = true);
    const char *ClassifyG(const char *, bool = true);
    const char *ClassifyH(const char *, bool = true);
    const char *ClassifyI(const char *, bool = true);
    const char *ClassifyL(const char *, bool = true);
    const char *ClassifyM(const char *, bool = true);
    const char *ClassifyN(const char *, bool = true);
    const char *ClassifyO(const char *, bool = true);
    const char *ClassifyP(const char *, bool = true);
    const char *ClassifyQ(const char *, bool = true);
    const char *ClassifyR(const char *, bool = true);
    const char *ClassifyS(const char *, bool = true);
    const char *ClassifyT(const char *, bool = true);
    const char *ClassifyV(const char *, bool = true);
    const char *ClassifyW(const char *, bool = true);
    const char *ClassifyX(const char *, bool = true);
    const char *ClassifyBadOption(const char *, bool = true);

    const char *ReportAmbiguousOption(const char *, const char *);
    const char *ReportMissingValue(const char *, const char *);
    const char *ReportValueNotRequired(const char *, const char *);

    void ProcessBlock(const char *, const char *, const char *);
    void ProcessHeader(const char *, const char *, const char *);
    void ProcessTrailer(const char *, const char *, const char *);
    void CheckBlockMarker(const char *, int);
    void CheckQuotedValue(const char *, const char *&, int &);

    void ProcessOptions(const char *);
    void ProcessPath(Tuple<const char *> &, const char *, const char * = NULL);
    const char *GetPrefix(const char *);
    const char *GetType(const char *);
};

#endif
