#include "parser.h"

#include <assert.h>
#include <iostream>
using namespace std;

#include "control.h"

Parser::Parser(Control *control_,
               LexStream *lex_stream_,
               VariableLookupTable *variable_table_,
               MacroLookupTable *macro_table_) : jikespg_act(control_,
                                                             lex_stream_,
                                                             variable_table_,
                                                             macro_table_)
{
    control = control_;
}

//
//
//
void Parser::Parse(int start_index)
{
    lex_stream -> Reset(start_index);

    curtok = lex_stream -> Gettoken();
    int act = START_STATE,
              current_kind = lex_stream -> Kind(curtok);

    //
    // Start parsing.
    //
    state_stack_top = -1;

    //
    // Process a terminal
    //
    for (;;)
    {
        if (++state_stack_top >= stack.Size())
            ReallocateStacks();

        stack[state_stack_top] = act;
        location_stack[state_stack_top] = Loc(curtok);

        act = t_action(act, current_kind, lex_stream);

        if (act <= NUM_RULES)
        {
            state_stack_top--; // make reduction look like a shift-reduce
        }
        else if (act > ERROR_ACTION)
        {
            curtok = lex_stream -> Gettoken();
            current_kind = lex_stream -> Kind(curtok);

            act -= ERROR_ACTION;
        }
        else if (act < ACCEPT_ACTION)
        {
            curtok = lex_stream -> Gettoken();
            current_kind = lex_stream -> Kind(curtok);

            continue;
        }
        else break;

        //
        // Process a non_terminal
        //
        do
        {
            state_stack_top -= (rhs[act] - 1);
            (this ->* rule_action[act])();
            act = nt_action(stack[state_stack_top], lhs[act]);
        } while(act <= NUM_RULES);
    } // process_termainal

    if (act == ERROR_ACTION)
    {
        //
        // Normally, we would want to do this:
        //
        //    DiagnoseParser *diagnose_parser = new DiagnoseParser(lex_stream);
        //    delete diagnose_parser;
        //

        int left_line_no    = lex_stream -> Line(curtok),
            left_column_no  = lex_stream -> Column(curtok),
            right_line_no   = lex_stream -> EndLine(curtok),
            right_column_no = lex_stream -> EndColumn(curtok),
            left_location   = lex_stream -> StartLocation(curtok),
            right_location  = lex_stream -> EndLocation(curtok);


        option -> report.Put(lex_stream -> FileName(curtok));
        option -> report.PutChar(':');
        option -> report.Put(left_line_no);
        option -> report.PutChar(':');
        option -> report.Put(left_column_no);
        option -> report.PutChar(':');
        option -> report.Put(right_line_no);
        option -> report.PutChar(':');
        option -> report.Put(right_column_no);
        option -> report.PutChar(':');
        option -> report.Put(left_location);
        option -> report.PutChar(':');
        option -> report.Put(right_location);
        option -> report.Put(": ");
        int  prevtok = lex_stream -> Previous(curtok);
        if (lex_stream -> Kind(prevtok) == TK_NOTICE_KEY ||
            lex_stream -> Kind(prevtok) == TK_GLOBALS_KEY ||
            lex_stream -> Kind(prevtok) == TK_AST_KEY  ||
            lex_stream -> Kind(prevtok) == TK_HEADERS_KEY)
        {
             BlockSymbol *block = option -> DefaultBlock();
             option -> report.Put("A block ");
             option -> report.Put(block -> BlockBegin());
             option -> report.Put("  ...  ");
             option -> report.Put(block -> BlockEnd());
             option -> report.Put(" is expected here instead of this token\n");
        }
        else option -> report.Put("Syntax error detected on this token\n");

        option -> report.Print(Control::syslis);
        option -> report.Flush(stdout);

        exit(12);
    }

    return;
}
