// XMOS
package org.eclipse.cdt.debug.mi.core.output;

/**
 * GDB/MI symbol list variables parsing.
 * -symbol-list-variables 1
 * ^done,locals=[{name="p",value="0x8048600 \"ghislaine\""},{name="buf",value="\"'\", 'x' <repeats 24 times>, \"i,xxxxxxxxx\", 'a' <repeats 24 times>"},{name="buf2",value="\"\\\"?'\\\\()~\""},{name="buf3",value="\"alain\""},{name="buf4",value="\"\\t\\t\\n\\f\\r\""},{name="i",value="0"}]
 *
 * On MacOS X 10.4 this returns a tuple:
 * ^done,locals={{name="p",value="0x8048600 \"ghislaine\""},{name="buf",value="\"'\", 'x' <repeats 24 times>, \"i,xxxxxxxxx\", 'a' <repeats 24 times>"},{name="buf2",value="\"\\\"?'\\\\()~\""},{name="buf3",value="\"alain\""},{name="buf4",value="\"\\t\\t\\n\\f\\r\""},{name="i",value="0"}}
 */
public class MISymbolListVariablesInfo extends MIInfo {

    MIArg[] globals;

    public MISymbolListVariablesInfo(MIOutput out) {
        super(out);
        parse();
    }

    public MIArg[] getLocals() {
        if (globals == null) {
            parse();
        }
        return globals;
    }

    void parse() {
        if (isDone()) {
            MIOutput out = getMIOutput();
            MIResultRecord rr = out.getMIResultRecord();
            if (rr != null) {
                MIResult[] results =  rr.getMIResults();
                for (int i = 0; i < results.length; i++) {
                    String var = results[i].getVariable();
                    if (var.equals("globals")) { //$NON-NLS-1$
                        MIValue value = results[i].getMIValue();
                        if (value instanceof MIList) {
                            globals = MIArg.getMIArgs((MIList)value);
                        } else if (value instanceof MITuple) {
                            globals = MIArg.getMIArgs((MITuple)value);
                        }
                    }
                }
            }
        }
        if (globals == null) {
            globals = new MIArg[0];
        }
    }
}
